"""
Created on Thu Oct 12 2017

@author: Giuseppe Armenise
"""

from warnings import warn

import numpy as np
import scipy as sc
from numpy.linalg import pinv

from .functionset import information_criterion, rescale
from .functionsetSIM import (
    K_calc,
    SS_lsim_process_form,
    Vn_mat,
    Z_dot_PIort,
    check_types,
    impile,
    ordinate_sequence,
    reducingOrder,
)
from .typing import ICMethods, OLSimMethods


def SVD_weighted(y, u, f, l_, weights: OLSimMethods = "N4SID"):
    Yf, Yp = ordinate_sequence(y, f, f)
    Uf, Up = ordinate_sequence(u, f, f)
    Zp = impile(Up, Yp)

    YfdotPIort_Uf = Z_dot_PIort(Yf, Uf)
    ZpdotPIort_Uf = Z_dot_PIort(Zp, Uf)
    O_i = np.dot(np.dot(YfdotPIort_Uf, pinv(ZpdotPIort_Uf)), Zp)

    if weights == "MOESP":
        W1 = None
        #        W2 = PIort_Uf
        OidotPIort_Uf = Z_dot_PIort(O_i, Uf)  # np.dot(O_i, W2)
        U_n, S_n, V_n = np.linalg.svd(OidotPIort_Uf, full_matrices=False)

    elif weights == "CVA":
        W1 = np.linalg.inv(
            sc.linalg.sqrtm(np.dot(YfdotPIort_Uf, YfdotPIort_Uf.T)).real
        )
        W1dotOi = np.dot(W1, O_i)
        W1_dot_Oi_dot_PIort_Uf = Z_dot_PIort(W1dotOi, Uf)
        U_n, S_n, V_n = np.linalg.svd(
            W1_dot_Oi_dot_PIort_Uf, full_matrices=False
        )

    elif weights == "N4SID":
        W1 = None  # is identity
        U_n, S_n, V_n = np.linalg.svd(
            O_i, full_matrices=False
        )  # full matrices not used

    return U_n, S_n, V_n, W1, O_i


def algorithm_1(
    y,
    u,
    l_,
    m_,
    f,
    N,
    U_n,
    S_n,
    V_n,
    W1,
    O_i,
    threshold,
    max_order,
    D_required,
):
    U_n, S_n, V_n = reducingOrder(U_n, S_n, V_n, threshold, max_order)
    V_n = V_n.T
    n = S_n.size
    S_n = np.diag(S_n)
    if W1 is None:  # W1 is identity
        Ob = np.dot(U_n, sc.linalg.sqrtm(S_n))
    else:
        Ob = np.dot(np.linalg.inv(W1), np.dot(U_n, sc.linalg.sqrtm(S_n)))
    X_fd = np.dot(np.linalg.pinv(Ob), O_i)
    Sxterm = impile(X_fd[:, 1:N], y[:, f : f + N - 1])
    Dxterm = impile(X_fd[:, 0 : N - 1], u[:, f : f + N - 1])
    if D_required:
        M = np.dot(Sxterm, np.linalg.pinv(Dxterm))
    else:
        M = np.zeros((n + l_, n + m_))
        M[0:n, :] = np.dot(Sxterm[0:n], np.linalg.pinv(Dxterm))
        M[n::, 0:n] = np.dot(Sxterm[n::], np.linalg.pinv(Dxterm[0:n, :]))
    residuals = Sxterm - np.dot(M, Dxterm)
    return Ob, X_fd, M, n, residuals


def forcing_A_stability(M, n, Ob, l_, X_fd, N, u, f):
    Forced_A = False
    if np.max(np.abs(np.linalg.eigvals(M[0:n, 0:n]))) >= 1.0:
        Forced_A = True
        print("Forcing A stability")
        M[0:n, 0:n] = np.dot(
            np.linalg.pinv(Ob), impile(Ob[l_::, :], np.zeros((l_, n)))
        )
        M[0:n, n::] = np.dot(
            X_fd[:, 1:N] - np.dot(M[0:n, 0:n], X_fd[:, 0 : N - 1]),
            np.linalg.pinv(u[:, f : f + N - 1]),
        )
    res = (
        X_fd[:, 1:N]
        - np.dot(M[0:n, 0:n], X_fd[:, 0 : N - 1])
        - np.dot(M[0:n, n::], u[:, f : f + N - 1])
    )
    return M, res, Forced_A


def extracting_matrices(M, n):
    A = M[0:n, 0:n]
    B = M[0:n, n::]
    C = M[n::, 0:n]
    D = M[n::, n::]
    return A, B, C, D


def OLSims(
    y: np.ndarray,
    u: np.ndarray,
    weights: OLSimMethods,
    order: int = 0,
    threshold: float = 0.0,
    f: int = 20,
    D_required: bool = False,
    A_stability: bool = False,
):
    y = np.atleast_2d(y)
    u = np.atleast_2d(u)
    l_, _ = y.shape
    m_, L = u.shape

    N = L - 2 * f + 1
    U_std = np.zeros(m_)
    Ystd = np.zeros(l_)
    for j in range(m_):
        U_std[j], u[j] = rescale(u[j])
    for j in range(l_):
        Ystd[j], y[j] = rescale(y[j])
    U_n, S_n, V_n, W1, O_i = SVD_weighted(y, u, f, l_, weights)
    Ob, X_fd, M, n, residuals = algorithm_1(
        y,
        u,
        l_,
        m_,
        f,
        N,
        U_n,
        S_n,
        V_n,
        W1,
        O_i,
        threshold,
        order,
        D_required,
    )
    if A_stability:
        M, residuals[0:n, :], _ = forcing_A_stability(
            M, n, Ob, l_, X_fd, N, u, f
        )
    A, B, C, D = extracting_matrices(M, n)
    Covariances = np.dot(residuals, residuals.T) / (N - 1)
    Q = Covariances[0:n, 0:n]
    R = Covariances[n::, n::]
    S = Covariances[0:n, n::]
    _, Y_estimate = SS_lsim_process_form(A, B, C, D, u)

    Vn = Vn_mat(y, Y_estimate)

    K, K_calculated = K_calc(A, C, Q, R, S)
    for j in range(m_):
        B[:, j] = B[:, j] / U_std[j]
        D[:, j] = D[:, j] / U_std[j]
    for j in range(l_):
        C[j, :] = C[j, :] * Ystd[j]
        D[j, :] = D[j, :] * Ystd[j]
        if K_calculated:
            K[:, j] = K[:, j] / Ystd[j]
    return A, B, C, D, Vn, Q, R, S, K


def select_order_SIM(
    y: np.ndarray,
    u: np.ndarray,
    weights: OLSimMethods,
    orders: tuple[int, int] = (1, 10),
    ic_method: ICMethods = "AIC",
    f: int = 20,
    D_required: bool = False,
    A_stability: bool = False,
):
    y = np.atleast_2d(y)
    u = np.atleast_2d(u)
    min_ord = min(orders)
    l_, L = y.shape
    m_, L = u.shape
    if not check_types(0.0, np.nan, np.nan, f):
        return (
            np.array([[0.0]]),
            np.array([[0.0]]),
            np.array([[0.0]]),
            np.array([[0.0]]),
            np.inf,
            [],
            [],
            [],
            [],
        )
    else:
        if min_ord < 1:
            warn("The minimum model order will be set to 1")
            min_ord = 1
        max_ord = max(orders) + 1
        if f < min_ord:
            warn(
                "The horizon must be larger than the model order, min_order set as f"
            )
            min_ord = f
        if f < max_ord - 1:
            warn(
                "The horizon must be larger than the model order, max_order set as f"
            )
            max_ord = f + 1
        IC_old = np.inf
        N = L - 2 * f + 1
        U_std = np.zeros(m_)
        Ystd = np.zeros(l_)
        for j in range(m_):
            U_std[j], u[j] = rescale(u[j])
        for j in range(l_):
            Ystd[j], y[j] = rescale(y[j])
        U_n, S_n, V_n, W1, O_i = SVD_weighted(y, u, f, l_, weights)
        for i in range(min_ord, max_ord):
            Ob, X_fd, M, n, residuals = algorithm_1(
                y, u, l_, m_, f, N, U_n, S_n, V_n, W1, O_i, 0.0, i, D_required
            )
            if A_stability:
                M, residuals[0:n, :], ForcedA = forcing_A_stability(
                    M, n, Ob, l_, X_fd, N, u, f
                )
                if ForcedA:
                    print("at n=", n)
                    print("--------------------")
            A, B, C, D = extracting_matrices(M, n)
            Covariances = np.dot(residuals, residuals.T) / (N - 1)
            X_states, Y_estimate = SS_lsim_process_form(A, B, C, D, u)

            Vn = Vn_mat(y, Y_estimate)

            K_par = n * l_ + m_ * n
            if D_required:
                K_par = K_par + l_ * m_
            IC = information_criterion(K_par, L, Vn, ic_method)
            if IC < IC_old:
                n_min = i
                IC_old = IC
        print("The suggested order is: n=", n_min)
        Ob, X_fd, M, n, residuals = algorithm_1(
            y, u, l_, m_, f, N, U_n, S_n, V_n, W1, O_i, 0.0, n_min, D_required
        )
        if A_stability:
            M, residuals[0:n, :], _ = forcing_A_stability(
                M, n, Ob, l_, X_fd, N, u, f
            )
        A, B, C, D = extracting_matrices(M, n)
        Covariances = np.dot(residuals, residuals.T) / (N - 1)
        X_states, Y_estimate = SS_lsim_process_form(A, B, C, D, u)

        Vn = Vn_mat(y, Y_estimate)

        Q = Covariances[0:n, 0:n]
        R = Covariances[n::, n::]
        S = Covariances[0:n, n::]
        K, K_calculated = K_calc(A, C, Q, R, S)
        for j in range(m_):
            B[:, j] = B[:, j] / U_std[j]
            D[:, j] = D[:, j] / U_std[j]
        for j in range(l_):
            C[j, :] = C[j, :] * Ystd[j]
            D[j, :] = D[j, :] * Ystd[j]
            if K_calculated:
                K[:, j] = K[:, j] / Ystd[j]
        return A, B, C, D, Vn, Q, R, S, K



"""
Created on Sun Oct 08 2017

@author: Giuseppe Armenise
"""

import math
from warnings import warn

import control.matlab as cnt
import numpy as np


def ordinate_sequence(y, f, p):
    l_, L = y.shape
    N = L - p - f + 1
    Yp = np.zeros((l_ * f, N))
    Yf = np.zeros((l_ * f, N))
    for i in range(1, f + 1):
        Yf[l_ * (i - 1) : l_ * i] = y[:, p + i - 1 : L - f + i]
        Yp[l_ * (i - 1) : l_ * i] = y[:, i - 1 : L - f - p + i]
    return Yf, Yp


def Z_dot_PIort(z, X):
    r"""
    Compute the scalar product between a vector z and $I - x^T \cdot pinv(X^T)$, avoiding the direct computation of the matrix

    PI = np.dot(X.T, np.linalg.pinv(X.T)), causing high memory usage


    Parameters
    ----------
    z : (...) vector array_like

    X : (...) matrix array_like

    """

    Z_dot_PIort = z - np.dot(np.dot(z, X.T), np.linalg.pinv(X.T))
    return Z_dot_PIort


def Vn_mat(y, yest):
    """
    Compute the variance of the model residuals

    Parameters
    ----------
    y : (L*l,1) vectorized matrix of output of the process

    yest : (L*l,1) vectorized matrix of output of the estimated model

    """
    y = y.flatten()
    yest = yest.flatten()
    eps = y - yest
    Vn = (eps @ eps) / (max(y.shape))  # @ is dot
    return Vn


def impile(M1, M2):
    M = np.zeros((M1[:, 0].size + M2[:, 0].size, M1[0, :].size))
    M[0 : M1[:, 0].size] = M1
    M[M1[:, 0].size : :] = M2
    return M


def reducingOrder(U_n, S_n, V_n, threshold=0.1, max_order=10):
    s0 = S_n[0]
    index = S_n.size
    for i in range(S_n.size):
        if S_n[i] < threshold * s0 or i >= max_order:
            index = i
            break
    return U_n[:, 0:index], S_n[0:index], V_n[0:index, :]


def check_types(threshold, max_order, fixed_order, f, p=20):
    if threshold < 0.0 or threshold >= 1.0:
        raise ValueError("The threshold value must be >=0. and <1.")
    if not np.isnan(max_order):
        if not isinstance(max_order, int):
            raise ValueError("The max_order value must be integer")
    if not np.isnan(fixed_order):
        if not isinstance(fixed_order, int):
            raise ValueError("The fixed_order value must be integer")
    if not isinstance(f, int):
        raise ValueError("The future horizon (f) must be integer")
    if not isinstance(p, int):
        raise ValueError("The past horizon (p) must be integer")
    return True


def check_inputs(threshold, max_order, fixed_order, f):
    if not math.isnan(fixed_order):
        threshold = 0.0
        max_order = fixed_order
    if f < max_order:
        warn(
            "The horizon must be larger than the model order, max_order setted as f"
        )
    if not max_order < f:
        max_order = f
    return threshold, max_order


def SS_lsim_process_form(A, B, C, D, u, x0=None):
    _, L = u.shape
    l_, n = C.shape
    y = np.zeros((l_, L))
    x = np.zeros((n, L))
    if x0 is not None:
        x[:, 0] = x0[:, 0]
    y[:, 0] = np.dot(C, x[:, 0]) + np.dot(D, u[:, 0])
    for i in range(1, L):
        x[:, i] = np.dot(A, x[:, i - 1]) + np.dot(B, u[:, i - 1])
        y[:, i] = np.dot(C, x[:, i]) + np.dot(D, u[:, i])
    return x, y


def SS_lsim_predictor_form(A_K, B_K, C, D, K, y, u, x0=None):
    _, L = u.shape
    l_, n = C.shape
    y_hat = np.zeros((l_, L))
    x = np.zeros((n, L + 1))
    if x0 is not None:
        x[:, 0] = x0[:, 0]
    for i in range(0, L):
        x[:, i + 1] = (
            np.dot(A_K, x[:, i]) + np.dot(B_K, u[:, i]) + np.dot(K, y[:, i])
        )
        y_hat[:, i] = np.dot(C, x[:, i]) + np.dot(D, u[:, i])
    return x, y_hat


def SS_lsim_innovation_form(A, B, C, D, K, y, u, x0=None):
    _, L = u.shape
    l_, n = C.shape
    y_hat = np.zeros((l_, L))
    x = np.zeros((n, L + 1))
    if x0 is not None:
        x[:, 0] = x0[:, 0]
    for i in range(0, L):
        y_hat[:, i] = np.dot(C, x[:, i]) + np.dot(D, u[:, i])
        x[:, i + 1] = (
            np.dot(A, x[:, i])
            + np.dot(B, u[:, i])
            + np.dot(K, y[:, i] - y_hat[:, i])
        )
    return x, y_hat


def K_calc(
    A: np.ndarray, C: np.ndarray, Q: np.ndarray, R: np.ndarray, S: np.ndarray
) -> tuple[np.ndarray, bool]:
    n_A = A[0, :].size
    try:
        P, _, _ = cnt.dare(A.T, C.T, Q, R, S, np.identity(n_A))
        K = np.dot(np.dot(A, P), C.T) + S
        K = np.dot(K, np.linalg.inv(np.dot(np.dot(C, P), C.T) + R))
        Calculated = True
    except ValueError:
        K = np.zeros((n_A, C[:, 0].size))
        warn("Kalman filter cannot be calculated")
        Calculated = False
    return K, Calculated



"""
Created on Wed Jul 26 2017

@author: Giuseppe Armenise
@updates: RBdC & MV
"""

from itertools import product
from typing import cast, get_args

import control.matlab as cnt
import numpy as np

from .armax import Armax, ARMAX_MISO_id
from .arx import ARX_id, ARX_MISO_id
from .functionset import information_criterion, rescale
from .io_opt import GEN_id, GEN_MISO_id
from .io_rls import GEN_RLS_id, GEN_RLS_MISO_id
from .typing import (
    Flags,
    ICMethods,
    IOMethods,
    OLSimMethods,
    OptMethods,
    PARSIMMethods,
    RLSMethods,
    SSMethods,
)
from .utils import (
    atleast_3d,
    check_feasibility,
    check_valid_orders,
    get_val_range,
)

n_orders = {
    "opt": ["na", "nb", "nc", "nd", "nf", "theta"],
    "rls": ["na", "nb", "nc", "nd", "nf", "theta"],
    "arx": ["na", "nb", "theta"],
    "armax": ["na", "nb", "nc", "theta"],
}


class SS_Model:
    def __init__(
        self,
        A,
        B,
        C,
        D,
        K,
        ts,
        Vn,
        x0=None,
        Q=None,
        R=None,
        S=None,
        A_K=None,
        B_K=None,
    ):
        self.A = A
        self.B = B
        self.C = C
        self.D = D
        self.Vn = Vn
        self.Q = Q  # not in parsim
        self.R = R  # not in parsim
        self.S = S  # not in parsim
        self.K = K

        self.n = A[:, 0].size
        self.G = cnt.ss(A, B, C, D, ts)
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = np.zeros((A[:, 0].size, 1))

        try:
            self.A_K = A - np.dot(K, C) if A_K is None else A_K
            self.B_K = B - np.dot(K, D) if B_K is None else B_K
        # TODO: find out why error is not raised
        except ValueError:
            self.A_K = []
            self.B_K = []

    @classmethod
    def _identify(
        cls,
        y: np.ndarray,
        u: np.ndarray,
        id_method: SSMethods,
        order: int | np.ndarray,
        f: int = 20,
        p: int = 20,
        threshold: float = 0.1,
        D_required: bool = False,
        A_stability: bool = False,
        B_recalc: bool = False,
    ):
        if isinstance(order, np.ndarray):
            order = int(order[0])
        if id_method in get_args(OLSimMethods):
            id_method = cast(OLSimMethods, id_method)
            from . import OLSims_methods

            A, B, C, D, Vn, Q, R, S, K = OLSims_methods.OLSims(
                y,
                u,
                id_method,
                order,
                threshold=threshold,
                f=f,
                D_required=D_required,
                A_stability=A_stability,
            )
            A_K, B_K, x0 = None, None, None
        else:
            id_method = cast(PARSIMMethods, id_method)
            from . import Parsim_methods

            A_K, C, B_K, D, K, A, B, x0, Vn = Parsim_methods.parsim(
                y,
                u,
                id_method,
                order,
                threshold,
                f,
                p,
                D_required,
                B_recalc,
            )
            Q, R, S = None, None, None

        return cls(A, B, C, D, K, 1.0, Vn, x0, Q, R, S, A_K, B_K)

    @classmethod
    def _from_order(
        cls,
        y: np.ndarray,
        u: np.ndarray,
        id_method: SSMethods,
        order: tuple[int, int],
        ic_method: ICMethods = "AIC",
        f: int = 20,
        p: int = 20,
        D_required: bool = False,
        A_stability: bool = False,
        B_recalc: bool = False,
    ):
        if id_method in get_args(OLSimMethods):
            id_method = cast(OLSimMethods, id_method)
            from . import OLSims_methods

            A, B, C, D, Vn, Q, R, S, K = OLSims_methods.select_order_SIM(
                y,
                u,
                id_method,
                order,
                ic_method,
                f,
                D_required,
                A_stability,
            )
            A_K, B_K, x0 = None, None, None
        else:
            id_method = cast(PARSIMMethods, id_method)
            from . import Parsim_methods

            A_K, C, B_K, D, K, A, B, x0, Vn = Parsim_methods.parsim(
                y,
                u,
                id_method,
                order,
                f=f,
                p=p,
                D_required=D_required,
                B_recalc=B_recalc,
                ic_method=ic_method,
            )
            Q, R, S = None, None, None

        return cls(A, B, C, D, K, 1.0, Vn, x0, Q, R, S, A_K, B_K)


class IO_SISO_Model:
    def __init__(
        self,
        G: cnt.TransferFunction,
        H: cnt.TransferFunction,
        numerator: np.ndarray,
        denominator: np.ndarray,
        numerator_H: np.ndarray,
        denominator_H: np.ndarray,
        *orders,
        Vn,
        Yid,
        **kwargs,
    ):
        self.G = G
        self.H = H

        self.numerator = numerator
        self.denominator = denominator
        self.numerator_H = numerator_H
        self.denominator_H = denominator_H

        self.orders = orders

        self.Vn = Vn
        self.Yid = Yid

        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def _from_order(
        cls,
        y: np.ndarray,
        u: np.ndarray,
        *ord_ranges: tuple[int, int],
        flag: Flags,
        id_method: IOMethods,
        ic_method: ICMethods = "AIC",
        ts: float = 1.0,
        max_iter: int = 200,
        stab_marg: float = 1.0,
        stab_cons: bool = False,
    ):
        if y.size != u.size:
            raise RuntimeError("y and u must have the same length")

        # order ranges
        ord_ranges_ = tuple(get_val_range(ord_r) for ord_r in ord_ranges)

        if ord_ranges_[1][0] <= 0:
            raise ValueError(
                f"Lower bound of nb must be strictly positive integer. Got {ord_ranges_[1][0]}"
            )

        y_std, y = rescale(y)
        U_std, u = rescale(u)
        IC_old = np.inf

        ord_range_best = tuple(ord_r[1] for ord_r in ord_ranges)

        for ord_range_prod in product(*ord_ranges):
            if flag == "opt":
                id_method = cast(OptMethods, id_method)
                _, _, _, _, Vn, Yid = GEN_id(
                    y,
                    u,
                    id_method,
                    *ord_range_prod,
                    max_iter=max_iter,
                    stab_marg=stab_marg,
                    stab_cons=stab_cons,
                )
            elif flag == "rls":
                id_method = cast(RLSMethods, id_method)
                _, _, _, _, Vn, Yid = GEN_RLS_id(
                    y, u, id_method, *ord_range_prod
                )
            elif flag == "arx":
                _, _, _, _, Vn, Yid = ARX_id(y, u, *ord_range_prod, y_std=1)
            elif flag == "armax":
                _, _, _, _, Vn, Yid = Armax._identify(
                    y,
                    u,
                    *ord_range_prod,
                    max_iter=max_iter,
                )

            IC = information_criterion(
                sum(ord_range_prod),
                y.size - max(ord_range_prod),
                Vn * 2,
                ic_method,
            )
            if IC < IC_old:
                IC_old = IC
                ord_range_best = ord_range_prod

        # rerun identification
        if flag == "opt":
            id_method = cast(OptMethods, id_method)
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                GEN_id(
                    y,
                    u,
                    id_method,
                    *ord_range_best,
                    max_iter=max_iter,
                    stab_marg=stab_marg,
                    stab_cons=stab_cons,
                )
            )
        elif flag == "rls":
            id_method = cast(RLSMethods, id_method)
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                GEN_RLS_id(y, u, id_method, *ord_range_best)
            )
        elif flag == "arx":
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                ARX_id(y, u, *ord_range_best, y_std=1.0)
            )
        elif flag == "armax":
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                Armax._identify(y, u, *ord_range_best, max_iter=max_iter)
            )

        Yid = np.atleast_2d(Yid) * y_std

        # rescale numerator coeff
        if id_method != "ARMA":
            nb = ord_range_best[1]
            theta = ord_range_best[-1]
            numerator[theta : nb + theta] = (
                numerator[theta : nb + theta] * y_std / U_std
            )

        # FdT
        G = cnt.tf(numerator, denominator, ts)
        H = cnt.tf(numerator_H, denominator_H, ts)

        if G is None or H is None:
            raise RuntimeError("tf could not be created")
        check_feasibility(G, H, id_method, stab_marg, stab_cons)

        return cls(
            G,
            H,
            numerator,
            denominator,
            numerator_H,
            denominator_H,
            *ord_range_best,
            Vn=Vn,
            Yid=Yid,
        )


class IO_MISO_Model(IO_SISO_Model):
    def __init__(
        self,
        G: cnt.TransferFunction,
        H: cnt.TransferFunction,
        numerator: np.ndarray,
        denominator: np.ndarray,
        numerator_H: np.ndarray,
        denominator_H: np.ndarray,
        *orders,
        Vn,
        Yid,
        **kwargs,
    ):
        super().__init__(
            G,
            H,
            numerator,
            denominator,
            numerator_H,
            denominator_H,
            *orders,
            Vn=Vn,
            Yid=Yid,
            **kwargs,
        )

    @classmethod
    def _identify(
        cls,
        y: np.ndarray,
        u: np.ndarray,
        flag: Flags,
        id_method: IOMethods,
        *orders: np.ndarray,
        ts: float,
        max_iter: int,
        stab_marg: float,
        stab_cons: bool,
        verbous: int = 0,
    ):
        udim = u.shape[0]
        na, nb, nc, nd, nf, theta = tuple(np.array(arg) for arg in orders)
        check_valid_orders(udim, *orders)

        # rerun identification
        if flag == "opt":
            id_method = cast(OptMethods, id_method)
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                GEN_MISO_id(
                    y,
                    u,
                    id_method,
                    int(na),
                    nb,
                    int(nc),
                    int(nd),
                    int(nf),
                    theta,
                    max_iter,
                    stab_marg,
                    stab_cons,
                )
            )
        elif flag == "rls":
            id_method = cast(RLSMethods, id_method)
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                GEN_RLS_MISO_id(
                    y,
                    u,
                    id_method,
                    int(na),
                    nb,
                    int(nc),
                    int(nd),
                    int(nf),
                    theta,
                )
            )
        elif flag == "arx":
            numerator, denominator, numerator_H, denominator_H, Vn, Yid = (
                ARX_MISO_id(y, u, int(na), nb, theta)
            )
        elif flag == "armax":
            numerator, denominator, numerator_H, denominator_H, Vn, Yid, _ = (
                ARMAX_MISO_id(
                    y,
                    u,
                    int(na),
                    nb,
                    int(nc),
                    theta,
                    max_iter,
                )
            )

        # FdT
        G = cnt.tf(atleast_3d(numerator), atleast_3d(denominator), ts)
        H = cnt.tf(atleast_3d(numerator_H), atleast_3d(denominator_H), ts)
        if G is None or H is None:
            raise RuntimeError("tf could not be created")

        check_feasibility(G, H, id_method, stab_marg, stab_cons)

        return cls(
            G,
            H,
            np.atleast_2d(numerator).tolist(),
            np.atleast_2d(denominator).tolist(),
            np.atleast_2d(numerator_H).tolist(),
            np.atleast_2d(denominator_H).tolist(),
            Vn=Vn,
            Yid=Yid,
        )


class IO_MIMO_Model(IO_MISO_Model):
    def __init__(
        self,
        G: cnt.TransferFunction,
        H: cnt.TransferFunction,
        numerator: np.ndarray,
        denominator: np.ndarray,
        numerator_H: np.ndarray,
        denominator_H: np.ndarray,
        *orders,
        Vn,
        Yid,
        **kwargs,
    ):
        super().__init__(
            G,
            H,
            numerator,
            denominator,
            numerator_H,
            denominator_H,
            *orders,
            Vn=Vn,
            Yid=Yid,
            **kwargs,
        )

    @classmethod
    def _identify(
        cls,
        y: np.ndarray,
        u: np.ndarray,
        flag: Flags,
        id_method: IOMethods,
        *orders: int | list[int] | np.ndarray,
        ts: float,
        max_iter: int,
        stab_marg: float,
        stab_cons: bool,
        verbous: int = 0,
    ):
        ydim, ylength = y.shape
        orders = tuple(np.array(arg) for arg in orders)
        check_valid_orders(ydim, *orders)
        # preallocation
        Vn_tot = 0.0
        numerator = []  # np.empty((ydim, u.shape[0], nbth?))
        denominator = []  # np.empty((ydim, u.shape[0], nbth?))
        numerator_H = []  # np.empty((ydim, u.shape[0], nbth?))
        denominator_H = []  # np.empty((ydim, u.shape[0], nbth?))
        Yid = np.zeros((ydim, ylength))
        # identification in MISO approach
        for i in range(ydim):
            miso_model = IO_MISO_Model._identify(
                y[i, :],
                u,
                flag,
                id_method,
                *(arg[i] for arg in orders),
                ts=ts,
                max_iter=max_iter,
                stab_marg=stab_marg,
                stab_cons=stab_cons,
            )
            # append values to vectors
            numerator.append(miso_model.numerator)
            denominator.append(miso_model.denominator)
            numerator_H.append(miso_model.numerator_H)
            denominator_H.append(miso_model.denominator_H)
            Vn_tot = miso_model.Vn + Vn_tot
            Yid[i, :] = miso_model.Yid

        if verbous == 1:
            print(f"Reached maximum iterations at output {i + 1}")
            print("-------------------------------------")

        # FdT
        G = cnt.tf(numerator, denominator, ts)
        H = cnt.tf(numerator_H, denominator_H, ts)
        if G is None or H is None:
            raise RuntimeError("tf could not be created")

        check_feasibility(G, H, id_method, stab_marg, stab_cons)

        return cls(
            G,
            H,
            numerator,
            denominator,
            numerator_H,
            denominator_H,
            *orders,
            Vn=Vn_tot,
            Yid=Yid,
        )



"""
Created on Wed Jul 26 2017

@author: Giuseppe Armenise
"""

import numpy as np

from .functionset import rescale
from .utils import check_valid_orders


def compute_phi(y, u, na, nb, theta, val, N, udim=1):
    """
    Compute the regressor matrix PHI.

    Parameters:
    y (np.ndarray): Output data.
    u (np.ndarray): Input data.
    na (int): Order of the autoregressive part.
    nb (int or np.ndarray): Order of the exogenous part.
    theta (int or np.ndarray): Delay of the exogenous part.
    val (int): Maximum predictable order.
    N (int): Number of data points.
    udim (int): Dimension of the input data.

    Returns:
    np.ndarray: Regressor matrix PHI.

    Examples:
    >>> y = np.array([1, 2, 3, 4, 5])
    >>> u = np.array([1, 2, 3, 4, 5])
    >>> na = 2
    >>> nb = np.array([2])
    >>> theta = np.array([1])
    >>> val = 2
    >>> N = 3
    >>> compute_phi(y, u, na, nb, theta, val, N, udim=1)
    array([[-2., -1.,  2.,  1.],
           [-3., -2.,  3.,  2.],
           [-4., -3.,  4.,  3.]])

    >>> y = np.array([1, 2, 3, 4, 5])
    >>> u = np.array([[1, 2, 3, 4, 5], [5, 4, 3, 2, 1]])
    >>> na = 2
    >>> nb = np.array([2, 2])
    >>> theta = np.array([1, 1])
    >>> val = 2
    >>> N = 3
    >>> compute_phi(y, u, na, nb, theta, val, N, udim=2)
    array([[-2., -1.,  1.,  1.,  5.,  5.],
           [-3., -2.,  2.,  1.,  4.,  5.],
           [-4., -3.,  3.,  2.,  3.,  4.]])
    """
    u = np.atleast_2d(u)
    phi = np.zeros(na + np.sum(nb[:]))
    PHI = np.zeros((N, na + np.sum(nb[:])))
    for k in range(N):
        phi[0:na] = -y[k + val - 1 :: -1][0:na]
        for nb_i in range(udim):
            phi[na + np.sum(nb[0:nb_i]) : na + np.sum(nb[0 : nb_i + 1])] = u[
                nb_i, :
            ][val + k - 1 :: -1][theta[nb_i] : nb[nb_i] + theta[nb_i]]
        PHI[k, :] = phi
    return PHI


def compute_theta(PHI, y, val, y_std=1.0):
    """
    Computes the parameter vector THETA, the model output y_id, and the estimated error norm Vn.
    Parameters:
    PHI (numpy.ndarray): The regression matrix.
    y (numpy.ndarray): The output vector.
    val (int): The index from which to start the validation.
    Returns:
    tuple: A tuple containing:
        - THETA (numpy.ndarray): The parameter vector.
        - y_id (numpy.ndarray): The model output including non-identified outputs.
        - Vn (float): The estimated error norm.
    Example:
    >>> import numpy as np
    >>> PHI = np.array([[1, 2], [3, 4], [5, 6]])
    >>> y = np.array([1, 2, 3, 4])
    >>> val = 1
    >>> THETA, y_id, Vn = compute_theta(PHI, y, val)
    >>> THETA
    array([-1. ,  1.5])
    >>> y_id
    array([1., 2., 3., 4.])
    >>> round(Vn, 6)
    np.float64(0.0)
    """

    # coeffiecients
    THETA = np.dot(np.linalg.pinv(PHI), y[val::])
    # model Output
    y_id0 = np.dot(PHI, THETA)
    # estimated error norm
    Vn = (np.linalg.norm((y_id0 - y[val::]), 2) ** 2) / (2 * (y.size - val))
    # adding non-identified outputs
    y_id = np.hstack((y[:val], y_id0))
    return THETA, y_id * y_std, Vn


def compute_num_den(
    THETA,
    na,
    nb,
    theta,
    val,
    udim: int = 1,
    y_std=1.0,
    U_std=np.array(1.0),
):
    """
    Compute the numerator and denominator coefficients.

    Parameters:
    THETA (np.ndarray): Coefficient vector.
    na (int): Order of the autoregressive part.
    nb (np.ndarray): Order of the exogenous part.
    theta (np.ndarray): Delay of the exogenous part.
    val (int): Maximum predictable order.
    udim (int): Dimension of the input data.
    y_std (float or np.ndarray): Standard deviation of the output data.
    U_std (float or np.ndarray): Standard deviation of the input data.

    Returns:
    tuple: Denominator coefficients, numerator coefficients, and NUMH.

    Examples:
    >>> THETA = np.array([0.5, -0.2, 0.3, 0.1])
    >>> na = 2
    >>> nb = np.array([2])
    >>> theta = np.array([1])
    >>> val = 3
    >>> udim = 1
    >>> compute_num_den(THETA, na, nb, theta, val, udim)
    (array([0. , 0.3, 0.1]), array([ 1. ,  0.5, -0.2,  0. ]))

    >>> THETA = np.array([0.5, -0.2, 0.3, 0.1, 0.4, 0.2])
    >>> na = 2
    >>> nb = np.array([2, 2])
    >>> theta = np.array([1, 1])
    >>> val = 3
    >>> udim = 2
    >>> compute_num_den(THETA, na, nb, theta, val, udim, y_std=1, U_std=[1.0, 1.0])
    (array([[0. , 0.3, 0.1],
           [0. , 0.4, 0.2]]), array([[ 1. ,  0.5, -0.2,  0. ],
           [ 1. ,  0.5, -0.2,  0. ]]))
    """
    NUM = np.zeros((udim, val))
    DEN = np.zeros((udim, val + 1))
    DEN[:, 0] = np.ones(udim)

    for k in range(udim):
        start = na + np.sum(nb[0:k])
        stop = na + np.sum(nb[0 : k + 1])
        THETA[start:stop] = THETA[start:stop] * y_std / np.atleast_1d(U_std)[k]
        NUM[k, theta[k] : theta[k] + nb[k]] = THETA[start:stop]
        DEN[k, 1 : na + 1] = THETA[0:na]

    return NUM.squeeze(), DEN.squeeze()


def ARX_id(
    y: np.ndarray,
    u: np.ndarray,
    na: int,
    nb: int | np.ndarray,
    theta: int | np.ndarray,
    y_std=1.0,
    U_std=np.array(1.0),
):
    # max predictable order
    val = max(na, np.max(nb + theta))
    N = y.size - val
    PHI = compute_phi(y, u, na, nb, theta, val, N, udim=1)

    THETA, y_id, Vn = compute_theta(PHI, y, val, y_std)

    NUM, DEN = compute_num_den(
        THETA, na, nb, theta, val, u.shape[0], y_std, U_std
    )
    NUMH = np.zeros_like(DEN)
    NUMH[0] = 1.0

    return NUM, DEN, NUMH, DEN, Vn, y_id


def ARX_MISO_id(
    y: np.ndarray,
    u: np.ndarray,
    na: int,
    nb: np.ndarray,
    theta: np.ndarray,
):
    nb = np.array(nb)
    theta = np.array(theta)
    u = np.atleast_2d(u)
    udim = u.shape[0]
    check_valid_orders(udim, *[nb, theta])

    y_std, y = rescale(y)
    U_std = np.zeros(udim)
    for j in range(udim):
        U_std[j], u[j] = rescale(u[j])

    return ARX_id(y, u, na, nb, theta, y_std, U_std)



"""
Created on Sun Sep 10 2017

@author: Giuseppe Armenise
"""

from warnings import warn

import control as cnt
import numpy as np

from .typing import CenteringMethods


# function which generates a sequence of inputs GBN
# N: sequence length (total number of samples)
# p_swd: desired probability of switching (no switch: 0<x<1 :always switch)
# Nmin: minimum number of samples between two switches
# Range: input range
# Tol: tolerance on switching probability relative error
# nit_max: maximum number of iterations
def GBN_seq(N, p_swd, Nmin=1, Range=[-1.0, 1.0], Tol=0.01, nit_max=30):
    min_Range = min(Range)
    max_Range = max(Range)
    prob = np.random.random()
    # set first value
    if prob < 0.5:
        gbn = -1.0 * np.ones(N)
    else:
        gbn = 1.0 * np.ones(N)
    # init. variables
    p_sw = p_sw_b = 2.0  # actual switch probability
    nit = 0
    while (np.abs(p_sw - p_swd)) / p_swd > Tol and nit <= nit_max:
        i_fl = 0
        Nsw = 0
        for i in range(N - 1):
            gbn[i + 1] = gbn[i]
            # test switch probability
            if i - i_fl >= Nmin:
                prob = np.random.random()
                # track last test of p_sw
                i_fl = i
                if prob < p_swd:
                    # switch and then count it
                    gbn[i + 1] = -gbn[i + 1]
                    Nsw = Nsw + 1
        # check actual switch probability
        p_sw = Nmin * (Nsw + 1) / N
        # set best iteration
        if np.abs(p_sw - p_swd) < np.abs(p_sw_b - p_swd):
            p_sw_b = p_sw
            Nswb = Nsw
            gbn_b = gbn.copy()
        # increase iteration number
        nit = nit + 1
    # rescale GBN
    for i in range(N):
        if gbn_b[i] > 0.0:
            gbn_b[i] = max_Range
        else:
            gbn_b[i] = min_Range
    return gbn_b, p_sw_b, Nswb


# function which generates a sequence of inputs as Random walk
# N: sequence length (total number of samples);
# sigma: standard deviation (mobility) of randow walk
# rw0: initial value
def RW_seq(N, rw0, sigma: float = 1.0):
    rw = rw0 * np.ones(N)
    for i in range(N - 1):
        # return random sample from a normal (Gaussian) distribution with:
        # mean = 0.0, standard deviation = sigma, and length = 1
        delta = np.random.normal(0.0, sigma, 1)
        # refresh input
        rw[i + 1] = (rw[i] + delta).item()
    return rw


# this function adds a white noise to a signal
# y:clean signal
# A_rel: relative amplitude (0<x<1) to the standard deviation of y (example: 0.05)
# noise amplitude=  A_rel*(standard deviation of y)
def white_noise(y, A_rel):
    num = y.size
    errors = np.zeros(num)
    y_err = np.zeros(num)
    Ystd = np.std(y)
    scale = np.abs(A_rel * Ystd)
    if scale < np.finfo(np.float32).eps:
        scale = np.finfo(np.float32).eps
        warn("A_rel may be too small, its value set to the lowest default one")

    errors = np.random.normal(0.0, scale, num)
    y_err = y + errors
    return errors, y_err


# this function generates a white noise matrix (rows with zero mean), L:size (columns), Var: variance vector
# e.g.   noise=white_noise_var(100,[1,1]) , noise matrix has two row vectors with variance=1
def white_noise_var(L, Var):
    Var = np.array(Var)
    n = Var.size
    noise = np.zeros((n, L))
    for i in range(n):
        if Var[i] < np.finfo(np.float32).eps:
            Var[i] = np.finfo(np.float32).eps
            warn(
                f"Var[{i}] may be too small, its value set to the lowest default one",
            )
        noise[i, :] = np.random.normal(0.0, Var[i] ** 0.5, L)
    return noise


# rescaling an array to its standard deviation. It gives the array rescaled: y=y/(standard deviation of y)
# and thestandard deviation: ex [Ystd,Y]=rescale(Y)
def rescale(y):
    y_std = float(np.std(y))
    y_scaled = y / y_std
    return y_std, y_scaled


def information_criterion(K, N, Variance, method="AIC"):
    if method == "AIC":
        IC = N * np.log(Variance) + 2 * K
    elif method == "AICc":
        if N - K - 1 > 0:
            IC = N * np.log(Variance) + 2 * K + 2 * K * (K + 1) / (N - K - 1)
        else:
            IC = np.inf
            raise RuntimeError(
                "Number of data is less than the number of parameters, AICc cannot be applied"
            )
    elif method == "BIC":
        IC = N * np.log(Variance) + K * np.log(N)
    return IC


def mean_square_error(predictions, targets):
    return ((predictions - targets) ** 2).mean()


# Function for model validation (one-step and k-step ahead predictor)
# SYS: system to validate (identified ARX or ARMAX model)
# u: input data
# y: output data
# Time: time sequence
# k: k-step ahead
def validation(SYS, u, y, Time, k=1, centering: CenteringMethods = None):
    # check dimensions
    y = np.atleast_2d(y)
    u = np.atleast_2d(u)
    [n1, n2] = y.shape
    ydim = min(n1, n2)
    ylength = max(n1, n2)
    if ylength == n1:
        y = y.T
    [n1, n2] = u.shape
    udim = min(n1, n2)
    ulength = max(n1, n2)
    if ulength == n1:
        u = u.T

    Yval = np.zeros((ydim, ylength))

    # Data centering
    if centering == "InitVal":
        y_rif = 1.0 * y[:, 0]
        u_rif = 1.0 * u[:, 0]
    elif centering == "MeanVal":
        for i in range(ydim):
            y_rif = np.mean(y, 1)
        for i in range(udim):
            u_rif = np.mean(u, 1)
    elif centering is None:
        y_rif = np.zeros(ydim)
        u_rif = np.zeros(udim)
    else:
        # elif centering != 'None':
        warn(
            "'Centering' argument is not valid, its value has been reset to 'None'"
        )

    # MISO approach
    # centering inputs and outputs
    for i in range(u.shape[0]):
        u[i, :] = u[i, :] - u_rif[i]
    for i in range(ydim):
        # one-step ahead predictor
        if k == 1:
            T, Y_u = cnt.forced_response(
                (1 / SYS.H[i, 0]) * SYS.G[i, :], Time, u
            )
            T, Y_y = cnt.forced_response(
                1 - (1 / SYS.H[i, 0]), Time, y[i, :] - y_rif[i]
            )
            Yval[i, :] = Y_u + np.atleast_2d(Y_y) + y_rif[i]
        else:
            # k-step ahead predictor
            # impulse response of disturbance model H
            T, hout = cnt.impulse_response(SYS.H[i, 0], T=Time)
            # extract first k-1 coefficients
            if hout is None:
                raise RuntimeError("H is not a valid transfer function")
            h_k_num = hout[0:k]
            # set denumerator
            h_k_den = np.hstack((np.ones((1, 1)), np.zeros((1, k - 1))))
            # FdT of impulse response
            Hk = cnt.tf(h_k_num, h_k_den[0], SYS.ts)
            # k-step ahead prediction
            T, Y_u = cnt.forced_response(
                Hk * (1 / SYS.H[i, 0]) * SYS.G[i, :], Time, u
            )
            T, Y_y = cnt.forced_response(
                1 - Hk * (1 / SYS.H[i, 0]), Time, y[i, :] - y_rif[i]
            )
            Yval[i, :] = np.atleast_2d(Y_u + Y_y + y_rif[i])

    return Yval



"""
Created on 2021

@author: RBdC & MV
"""

import numpy as np

from .functionset import rescale
from .typing import RLSMethods
from .utils import build_tfs, common_setup, validate_and_prepare_inputs

# ----------------- Helper Functions -----------------


def _initialize_parameters(
    N: int, nt: int, y: np.ndarray | None = None
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Initialize P_t, teta, eta, and Yp."""
    Beta = 1e4
    p_t = Beta * np.eye(nt, nt)
    P_t = np.repeat(p_t[:, :, np.newaxis], N, axis=2)
    teta = np.zeros((nt, N))
    eta = np.zeros(N)
    Yp = y.copy() if y is not None else np.zeros(N)
    return P_t, teta, eta, Yp


def _compute_error_norm(y: np.ndarray, Yp: np.ndarray, val: int) -> float:
    """Calculate the normalized prediction error."""
    return float(np.linalg.norm(y - Yp, 2) ** 2) / (2 * (y.size - val))


def _propagate_parameters(
    y,
    u,
    na,
    nb,
    nc,
    nd,
    nf,
    theta,
    id_method,
    val,
    P_t,
    teta,
    eta,
    Yp,
    nt,
):
    N = y.size
    # Gain
    K_t = np.zeros((nt, N))

    # Forgetting factors
    L_t = 1
    l_t = L_t * np.ones(N)
    #
    E = np.zeros(N)
    fi = np.zeros((1, nt, N))

    # Propagation
    for k in range(N):
        if k > val:
            # Step 1: Regressor vector
            vecY = y[k - na : k][::-1]  # Y vector
            vecYp = Yp[k - nf : k][::-1]  # Yp vector
            #
            # vecE = E[k-nh:k][::-1]                     # E vector

            vecU = []
            for nb_i in range(nb.size):  # U vector
                vecu = u[nb_i][k - nb[nb_i] - theta[nb_i] : k - theta[nb_i]][
                    ::-1
                ]
                vecU = np.hstack((vecU, vecu))  # U vector

                vecE = E[k - nc : k][::-1]

            # choose input-output model
            if id_method == "ARMAX":
                fi[:, :, k] = np.hstack((-vecY, vecU, vecE))
            elif id_method == "ARX":
                fi[:, :, k] = np.hstack((-vecY, vecU))
            elif id_method == "OE":
                fi[:, :, k] = np.hstack((-vecYp, vecU))
            elif id_method == "FIR":
                fi[:, :, k] = vecU
            phi = fi[:, :, k].T

            # Step 2: Gain Update
            # Gain of parameter teta
            K_t[:, k : k + 1] = np.dot(
                np.dot(P_t[:, :, k - 1], phi),
                np.linalg.inv(
                    l_t[k - 1] + np.dot(np.dot(phi.T, P_t[:, :, k - 1]), phi)
                ),
            )

            # Step 3: Parameter Update
            teta[:, k] = teta[:, k - 1] + np.dot(
                K_t[:, k : k + 1], (y[k] - np.dot(phi.T, teta[:, k - 1]))
            )

            # Step 4: A posteriori prediction-error
            Yp[k] = np.dot(phi.T, teta[:, k]).item() + eta[k]
            E[k] = y[k] - Yp[k]

            # Step 5. Parameter estimate covariance update
            P_t[:, :, k] = (1 / l_t[k - 1]) * (
                np.dot(
                    np.eye(nt) - np.dot(K_t[:, k : k + 1], phi.T),
                    P_t[:, :, k - 1],
                )
            )

            # Step 6: Forgetting factor update
            l_t[k] = 1.0
    return teta, Yp


# ----------------- Main Functions -----------------


def GEN_RLS_id(
    y: np.ndarray,
    u: np.ndarray,
    id_method: RLSMethods,
    na: int,
    nb: int | np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: int | np.ndarray,
    **_,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, np.ndarray]:
    # Input handling and validation
    u, nb, theta, udim = validate_and_prepare_inputs(u, nb, theta)
    val, nt = common_setup(na, nb, nc, nd, nf, theta)

    # Parameter initialization
    P_t, teta, eta, Yp = _initialize_parameters(y.size, nt)

    # Propagate parameters
    teta, Yp = _propagate_parameters(
        y, u, na, nb, nc, nd, nf, theta, id_method, val, P_t, teta, eta, Yp, nt
    )

    # Compute results
    Vn = _compute_error_norm(y, Yp, val)
    THETA = teta[:, -1]
    NUM, DEN, NUMH, DENH = build_tfs(
        THETA, na, nb, nc, nd, nf, theta, id_method, udim
    )

    return NUM.squeeze(), DEN.squeeze(), NUMH.squeeze(), DENH.squeeze(), Vn, Yp


def GEN_RLS_MISO_id(
    y: np.ndarray,
    u: np.ndarray,
    id_method: RLSMethods,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, float, np.ndarray]:
    # Input handling and scaling
    u, nb, theta, udim = validate_and_prepare_inputs(u, nb, theta)
    y_std, y = rescale(y)
    U_std = np.zeros(udim)
    for j in range(udim):
        U_std[j], u[j] = rescale(u[j])

    # Common setup
    val, nt = common_setup(na, nb, nc, nd, nf, theta)
    P_t, teta, eta, Yp = _initialize_parameters(y.size, nt, y)

    # Propagate parameters
    teta, Yp = _propagate_parameters(
        y,
        u,
        na,
        nb,
        nc,
        nd,
        nf,
        theta,
        id_method,
        val,
        P_t,
        teta,
        eta,
        Yp,
        nt,
    )

    # Compute results
    Vn = _compute_error_norm(y, Yp, val)
    THETA = teta[:, -1]

    # Adjust coefficients for scaling
    start_idx = na if id_method != "OE" else nf
    for k in range(udim):
        coeffs = THETA[
            start_idx + np.sum(nb[:k]) : start_idx + np.sum(nb[: k + 1])
        ]
        THETA[start_idx + np.sum(nb[:k]) : start_idx + np.sum(nb[: k + 1])] = (
            coeffs * y_std / U_std[k]
        )

    NUM, DEN, NUMH, DENH = build_tfs(
        THETA, na, nb, nc, nd, nf, theta, id_method, udim, y_std, U_std
    )
    y_id = Yp * y_std

    return NUM, DEN, NUMH, DENH, Vn, y_id



"""
Created on 2017

@author: Giuseppe Armenise

@updates: Riccardo Bacci di Capaci, Marco Vaccari, Federico Pelagagge
"""

from collections.abc import Mapping
from typing import cast, get_args, overload
from warnings import warn

import numpy as np

from .model import IO_MIMO_Model, IO_SISO_Model, SS_Model
from .typing import (
    AvailableMethods,
    AvailableModes,
    CenteringMethods,
    Flags,
    ICMethods,
    IOMethods,
    SSMethods,
)

ID_MODES: dict[AvailableModes, Flags] = {
    "LLS": "arx",
    "RLLS": "rls",
    "OPT": "opt",
    "ILLS": "armax",
}

METHOD_ORDERS: dict[AvailableMethods, list[str]] = {
    "FIR": ["na", "nb", "theta"],
    "ARX": ["na", "nb", "theta"],
    "ARMAX": ["na", "nb", "nc", "theta"],
    "OE": ["nb", "nf", "theta"],
    "ARMA": ["na", "nc", "theta"],
    "ARARX": ["na", "nb", "nd", "theta"],
    "ARARMAX": ["na", "nb", "nc", "nd", "theta"],
    "BJ": ["nb", "nc", "nd", "nf", "theta"],
    "GEN": ["na", "nb", "nc", "nd", "nf", "theta"],
    "EARMAX": ["na", "nb", "nc", "theta"],
    "EOE": ["nb", "nf", "theta"],
    "CVA": ["na"],
    "MOESP": ["na"],
    "N4SID": ["na"],
    "PARSIM_K": ["na"],
    "PARSIM_P": ["na"],
    "PARSIM_S": ["na"],
}


@overload
def _as_orders_defaults(
    orders_dict: Mapping[str, int | list | np.ndarray],
    orders_defaults: Mapping[str, np.ndarray],
) -> Mapping[str, np.ndarray]: ...
@overload
def _as_orders_defaults(
    orders_dict: Mapping[str, tuple[int, int]],
    orders_defaults: Mapping[str, tuple[int, int]],
) -> Mapping[str, tuple[int, int]]: ...


def _as_orders_defaults(
    orders_dict: Mapping[str, int | list | np.ndarray | tuple[int, int]],
    orders_defaults: Mapping[str, np.ndarray | tuple[int, int]],
) -> Mapping[str, np.ndarray | tuple[int, int]]:
    """
    Ensure that the orders dictionary has the correct shape and type.

    Parameters:
    orders (Mapping[str, int | list | np.ndarray]): The orders to check.
    orders_defaults (Mapping[str, np.ndarray]): The default orders to use for shape reference.

    Returns:
    dict[str, np.ndarray]: The validated and fixed orders.

    Raises:
    RuntimeError: If the order is not an int, list, or np.ndarray, or if the list length does not match the default shape.

    Examples:
    >>> orders_defaults = {'na': np.zeros((1,)), 'nb': np.zeros((2,2)), 'nc': np.zeros((2,))}
    >>> orders = {'na': 2, 'nb': [[1, 2], [3, 4]], 'nc': np.array([3, 4])}
    >>> _as_orders_defaults(orders, orders_defaults)
    {'na': array([2]), 'nb': array([[1, 2], [3, 4]]), 'nc': array([3, 4])}

    >>> orders = {'na': 2, 'nb': [1, 2, 3], 'nc': np.array([3, 4])}
    >>> _as_orders_defaults(orders, orders_defaults)
    Traceback (most recent call last):
        ...
    RuntimeError: Order for nb must have 2 elements

    >>> orders_defaults = {'na': (0, 0), 'nb': (0, 0), 'nc': (0, 0)}
    >>> orders = {'na': (0, 0), 'nb': (0, 0), 'nc': (0, 0)}
    >>> _as_orders_defaults(orders, orders_defaults)
    {'na': (0, 0), 'nb': (0, 0), 'nc': (0, 0)}
    >>> orders = {'na': 0, 'nb': (0, 0), 'nc': (0, 0)}
    >>> _as_orders_defaults(orders, orders_defaults)
    Traceback (most recent call last):
        ...
    RuntimeError: Order for na must be convertible to (0, 0). Got 0 instead.
    """
    orders_: dict[str, np.ndarray | tuple[int, int]] = {}
    for name, order in orders_dict.items():
        order_defaults = orders_defaults[name]
        if isinstance(order_defaults, np.ndarray):
            shape = order_defaults.shape
            if isinstance(order, int):
                order_ = order * np.ones(shape, dtype=int)
            elif isinstance(order, list | np.ndarray):
                order_ = np.array(order, dtype=int)
                if order_.shape != shape:
                    raise RuntimeError(
                        f"Order for {name} must be of shape {shape}. Got {order_.shape} instead."
                    )
            orders_[name] = order_
        elif isinstance(order, tuple):
            if len(order) != 2:
                raise RuntimeError(
                    f"Order for {name} must have 2 elements. Got {len(order)} instead."
                )
            orders_[name] = order

        else:
            raise RuntimeError(
                f"Order for {name} must be convertible to {order_defaults}. Got {order} instead."
            )

    return orders_


def _areinstances(args: tuple, class_or_tuple):
    """
    Check if all arguments are instances of a given class or tuple of classes.
    Args:
        *args: Variable length argument list of objects to check.
        class_or_tuple (type or tuple of types): The class or tuple of classes to check against.
    Returns:
        bool: True if all arguments are instances of the given class or tuple of classes, False otherwise.
    Examples:
        >>> _areinstances((1, 2, 3), int)
        True
        >>> _areinstances((1, 'a', 3), int)
        False
        >>> _areinstances((1, 'a', 3), (int, str))
        True
    """

    return all(isinstance(x, class_or_tuple) for x in args)


def _verify_orders_types(
    *orders: int | list | np.ndarray | tuple, IC: ICMethods | None = None
):
    """
    Validates the orders types.

    Args:
        orders (list): A list of orders which can be of type int, list, or tuple.
        IC (ICMethods | None): An instance of ICMethods or None.

    Raises:
        ValueError: If orders are tuples and IC is not one of the ICMethods.
        ValueError: If orders are not all int, list, or tuple.

    Examples:
        >>> _verify_orders_types(*[1, 2, 3])
        >>> _verify_orders_types(*[(1, 2), (3, 4)], IC="AIC")
        >>> _verify_orders_types(*[1, [2, 3], (4, 5)])
        Traceback (most recent call last):
        ...
        ValueError: All orders must be either int, list, or tuple.Got [<class 'int'>, <class 'list'>, <class 'tuple'>] instead.
        >>> _verify_orders_types(*[(1, 2), (3, 4)])
        Traceback (most recent call last):
        ...
        ValueError: IC must be one of ('AIC', 'AICc', ...) if orders are tuples.
    """
    if _areinstances(orders, tuple):
        if IC is None or IC not in get_args(ICMethods):
            raise ValueError(
                f"IC must be one of {get_args(ICMethods)} if orders are tuples. Got {IC} with {orders} instead."
            )
    elif not (_areinstances(orders, int) or _areinstances(orders, list)):
        raise ValueError(
            "All orders must be either int, list, or tuple."
            f"Got {[type(order) for order in orders]} instead."
        )


def _verify_orders_len(
    id_method: AvailableMethods,
    *orders: int | list | np.ndarray | tuple,
    ydim: int,
    IC: ICMethods | None,
):
    """
    Verify the number and values of orders for a given identification method.

    Args:
        id_method (AvailableMethods): The identification method to be used.
        *orders (int | list | np.ndarray | tuple): Variable length argument list of orders.
        ydim (int): The dimension of the output.
        IC (ICMethods | None): The information criterion method, if any.

    Raises:
        ValueError: If the number of orders does not match the required number of orders for the method.
        ValueError: If the order 'na' for FIR is not valid.

    Examples:
        No exception raised
        >>> _verify_orders_len("FIR", *[[0,0], [1, 2], [2,3]], ydim=2, IC=None)

        >>> _verify_orders_len("FIR", *[1, 0], ydim=2, IC=None)
        Traceback (most recent call last):
            ...
        ValueError: Order 'na' for FIR must be [0, 0]. Got [1, 0] instead.

        >>> _verify_orders_len("ARX", 1, 2, ydim=2, IC=None)
        Traceback (most recent call last):
            ...
        ValueError: Number of orders (2) does not match the number of required orders (3). Required are [na, nb, theta] got [1, 2]
    """
    method_orders = METHOD_ORDERS[id_method]
    # TODO: allow user to not define `na` for FIR.
    if id_method == "FIR":
        if orders[0] != [0] * ydim and orders[0] != (0, 0):
            raise ValueError(
                f"Order 'na' for FIR must be {[0] * ydim if IC is None or IC not in get_args(ICMethods) else (0, 0)}. Got {orders[0]} instead."
            )

    if len(orders) != len(method_orders):
        raise ValueError(
            f"Number of orders ({len(orders)}) does not match the number of required orders ({len(method_orders)})."
            f"Required are {method_orders} got [{', '.join(map(str, orders))}]"
        )


@overload
def _update_orders(
    orders: tuple[int | list[int] | list[list[int]] | np.ndarray, ...],
    orders_defaults: Mapping[str, np.ndarray],
    id_method: AvailableMethods,
) -> tuple[np.ndarray, ...]: ...
@overload
def _update_orders(
    orders: tuple[tuple[int, int], ...],
    orders_defaults: Mapping[str, tuple[int, int]],
    id_method: AvailableMethods,
) -> tuple[tuple[int, int], ...]: ...


def _update_orders(
    orders: tuple[
        int | list[int] | list[list[int]] | np.ndarray | tuple[int, int], ...
    ],
    orders_defaults: Mapping[str, np.ndarray | tuple[int, int]],
    id_method: AvailableMethods,
) -> tuple[np.ndarray | tuple[int, int], ...]:
    """
    Consolidates two dictionaries of orders, giving precedence to the values in `orders`.
    This function merges `orders_defaults` and `orders`, with `orders` values
    taking precedence over `orders_defaults`. It then checks and fixes the consolidated
    orders using the `_as_orders_defaults` function and returns the final orders as a tuple.
    Args:
        orders: The orders dictionary with updated values.
        orders_defaults: The default orders dictionary.
    Returns:
        tuple: A tuple containing the consolidated orders.
    Examples:
        >>> orders = ([0], [[1, 2], [3, 4]], np.array([[1, 2], [3, 4]]))
        >>> orders_defaults = {'na': np.zeros((1,)), 'nb': np.zeros((2,2)), 'nc': np.zeros((2,)), 'nd': np.zeros((2,)), 'nf': np.zeros((2,)), 'theta': np.zeros((2,2))}
        >>> _update_orders(orders, orders_defaults, id_method="FIR")
        (array([0]), array([[1, 2], [3, 4]]), array([0, 0]), array([0, 0]), array([0, 0]), array([[1, 2], [3, 4]]))

        >>> orders = (2, [1, 2, 3], np.array([3, 4]))
        >>> _update_orders(orders, orders_defaults, id_method="FIR")
        Traceback (most recent call last):
            ...
        RuntimeError: Order for nb must be of shape (2, 2). Got (3,) instead.
    """
    orders_dict = dict(zip(METHOD_ORDERS[id_method], orders, strict=True))
    orders_up = dict(orders_defaults)
    orders_up.update(orders_dict)
    orders_up = _as_orders_defaults(orders_up, orders_defaults)
    return tuple(orders_up.values())


def _recentering_transform(y, y_rif):
    ylength = y.shape[1]
    for i in range(ylength):
        y[:, i] = y[:, i] + y_rif
    return y


def _recentering_fit_transform(y, u, centering: CenteringMethods = None):
    ydim, ylength = y.shape
    udim, ulength = u.shape
    if centering == "InitVal":
        y_rif = 1.0 * y[:, 0]
        u_init = 1.0 * u[:, 0]
        for i in range(ylength):
            y[:, i] = y[:, i] - y_rif
            u[:, i] = u[:, i] - u_init
    elif centering == "MeanVal":
        y_rif = np.zeros(ydim)
        u_mean = np.zeros(udim)
        for i in range(ydim):
            y_rif[i] = np.mean(y[i, :])
        for i in range(udim):
            u_mean[i] = np.mean(u[i, :])
        for i in range(ylength):
            y[:, i] = y[:, i] - y_rif
            u[:, i] = u[:, i] - u_mean
    else:
        if centering is not None:
            warn(
                "'centering' argument is not valid, its value has been reset to 'None'"
            )
        y_rif = 0.0 * y[:, 0]

    return y, u, y_rif


# TODO: learn how to provide overloads without extensively typing out irrelevant args
# @overload
# def system_identification(
#     y: np.ndarray,
#     u: np.ndarray,
#     id_method: AvailableMethods,
#     *orders: int | list[int] | list[list[int]] | np.ndarray,
#     IC: None = None,
# ): ...
# @overload
# def system_identification(
#     y: np.ndarray,
#     u: np.ndarray,
#     id_method: AvailableMethods,
#     *orders: tuple[int, int],
#     IC: ICMethods,
# ): ...


def system_identification(
    y: np.ndarray,
    u: np.ndarray,
    id_method: AvailableMethods,
    *orders: int | list[int] | list[list[int]] | np.ndarray | tuple[int, int],
    ts: float = 1.0,
    centering: CenteringMethods | None = None,
    IC: ICMethods | None = None,
    id_mode: AvailableModes = "OPT",  # TODO: figure out whether to remove default
    max_iter: int = 200,
    stab_marg: float = 1.0,
    stab_cons: bool = False,
    SS_f: int = 20,
    SS_p: int = 20,
    SS_threshold: float = 0.1,
    SS_D_required: bool = False,
    SS_A_stability: bool = False,
    SS_PK_B_reval: bool = False,
) -> IO_SISO_Model | IO_MIMO_Model | SS_Model:
    # Verify y and u
    y = np.atleast_2d(y)
    u = np.atleast_2d(u)
    [n1, n2] = y.shape
    ydim = min(n1, n2)
    ylength = max(n1, n2)
    if ylength == n1:
        y = y.T
    [n1, n2] = u.shape
    ulength = max(n1, n2)
    udim = min(n1, n2)
    if ulength == n1:
        u = u.T

    # Checking data consinstency
    if ulength != ylength:
        warn(
            "y and u lengths are not the same. The minor value between the two lengths has been chosen. The perfomed indentification may be not correct, be sure to check your input and output data alignement"
        )
        # Recasting data cutting out the over numbered data
        minlength = min(ulength, ylength)
        y = y[:, :minlength]
        u = u[:, :minlength]

    _verify_orders_types(*orders, IC=IC)

    _verify_orders_len(id_method, *orders, ydim=ydim, IC=IC)

    # Data centering
    y, u, y_rif = _recentering_fit_transform(y, u, centering)

    ##### Check Information Criterion #####

    # MODE 1) fixed orders
    if not _areinstances(orders, tuple):
        if IC is not None:
            warn("Ignoring argument 'IC' as fixed orders are provided.")

        orders = cast(
            tuple[int | list[int] | list[list[int]] | np.ndarray, ...], orders
        )
        orders_defaults: Mapping[str, np.ndarray] = {
            "na": np.zeros((ydim,), dtype=int),
            "nb": np.zeros((ydim, udim), dtype=int),
            "nc": np.zeros((ydim,), dtype=int),
            "nd": np.zeros((ydim,), dtype=int),
            "nf": np.zeros((ydim,), dtype=int),
            "theta": np.zeros((ydim, udim), dtype=int),
        }
        orders = _update_orders(orders, orders_defaults, id_method=id_method)

        # IO Models
        if id_method in get_args(IOMethods):
            id_method = cast(IOMethods, id_method)
            if id_mode in get_args(AvailableModes):
                flag = ID_MODES[id_mode]
            else:
                raise RuntimeError(
                    f"Method {id_mode} not available for {id_method}. Available: {get_args(AvailableModes)}"
                )

            model = IO_MIMO_Model._identify(
                y,
                u,
                flag,
                id_method,
                *orders,
                ts=ts,
                max_iter=max_iter,
                stab_marg=stab_marg,
                stab_cons=stab_cons,
            )

            model.Yid = _recentering_transform(model.Yid, y_rif)

        # SS MODELS
        elif id_method in get_args(SSMethods):
            id_method = cast(SSMethods, id_method)
            order = orders[0]
            model = SS_Model._identify(
                y,
                u,
                id_method,
                order,
                SS_f,
                SS_p,
                SS_threshold,
                SS_D_required,
                SS_PK_B_reval,
            )

        # NO method selected
        else:
            raise RuntimeError(
                f"Wrong identification method selected. Got {id_method}"
                f"expected one of {get_args(AvailableMethods)}"
            )

    # =========================================================================
    # MODE 2) order range
    # if an IC is selected
    else:
        IC = cast(ICMethods, IC)
        if ydim != 1 or udim != 1:
            raise RuntimeError(
                "Information criteria are implemented ONLY in SISO case "
                "for INPUT-OUTPUT model sets. Use subspace methods instead"
                " for MIMO cases"
            )

        orders = cast(tuple[tuple[int, int], ...], orders)
        orders_ranges_defaults: Mapping[str, tuple[int, int]] = {
            "na": (0, 0),
            "nb": (0, 0),
            "nc": (0, 0),
            "nd": (0, 0),
            "nf": (0, 0),
            "theta": (0, 0),
        }
        orders = _update_orders(
            orders, orders_ranges_defaults, id_method=id_method
        )

        if id_method in get_args(IOMethods):
            id_method = cast(IOMethods, id_method)
            if id_mode in get_args(AvailableModes):
                flag = ID_MODES[id_mode]
            else:
                raise RuntimeError(
                    f"Method {id_mode} not available for {id_method}. Available: {get_args(AvailableModes)}"
                )

            model = IO_SISO_Model._from_order(
                y[0],
                u[0],
                *orders,
                flag=flag,
                id_method=id_method,
                ts=ts,
                ic_method=IC,
                max_iter=max_iter,
                stab_marg=stab_marg,
                stab_cons=stab_cons,
            )
            model.Yid = _recentering_transform(model.Yid, y_rif)

        # SS-MODELS
        elif id_method in get_args(SSMethods):
            id_method = cast(SSMethods, id_method)
            order = orders[0]
            model = SS_Model._from_order(
                y,
                u,
                id_method,
                order,
                IC,
                SS_f,
                SS_p,
                SS_D_required,
                SS_PK_B_reval,
            )

        # NO method selected
        else:
            raise RuntimeError(
                f"Wrong identification method selected. Got {id_method}"
                f"expected one of {get_args(AvailableMethods)}"
            )

    return model



"""
Created on 2021

@author: RBdC & MV
"""

from warnings import warn

import numpy as np

from .functionset import rescale
from .functionset_OPT import opt_id
from .typing import IOMethods, OptMethods
from .utils import (
    build_tfs,
    check_valid_orders,
    common_setup,
    validate_and_prepare_inputs,
)

# ----------------- Helper Functions -----------------


def _build_initial_guess(y, n_coeff: int, id_method: IOMethods) -> np.ndarray:
    w_0 = np.zeros((1, n_coeff))
    w_y = np.atleast_2d(y)
    w_0 = np.hstack([w_0, w_y])
    if id_method in ["BJ", "GEN", "ARARX", "ARARMAX"]:
        w_0 = np.hstack([w_0, w_y, w_y])
    return w_0


def _extract_results(
    sol, n_coeff: int, ylength: int, y_std: float = 1.0
) -> tuple:
    x_opt = sol["x"]
    THETA = np.array(x_opt[:n_coeff])[:, 0]
    y_id0 = x_opt[-ylength:].full()[:, 0]
    y_id = y_id0 * y_std
    return THETA, y_id


# ----------------- Main Functions -----------------


def GEN_id(
    y: np.ndarray,
    u: np.ndarray,
    id_method: OptMethods,
    na: int,
    nb: int | np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: int | np.ndarray,
    max_iter: int,
    stab_marg: float,
    stab_cons: bool,
    adjust_B: bool = False,
    y_std: float = 1.0,
    U_std: np.ndarray = np.array([1.0]),
) -> tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.floating, np.ndarray
]:
    u, nb, theta, udim = validate_and_prepare_inputs(u, nb, theta)
    val, n_coeff = common_setup(na, nb, nc, nd, nf, theta)
    solver, w_lb, w_ub, g_lb, g_ub = opt_id(
        y,
        u,
        id_method,
        na,
        nb,
        nc,
        nd,
        nf,
        theta,
        max_iter,
        stab_marg,
        stab_cons,
        n_coeff,
        udim,
        val,
    )
    iterations = solver.stats()["iter_count"]
    if iterations >= max_iter:
        warn("Reached maximum number of iterations")

    w_0 = _build_initial_guess(y, n_coeff, id_method)
    sol = solver(lbx=w_lb, ubx=w_ub, x0=w_0, lbg=g_lb, ubg=g_ub)

    THETA, y_id = _extract_results(sol, n_coeff, y.size)
    y_id = y_id * y_std
    # TODO: this is currently implemented within build_tf_G()
    # Adjust B coefficients with scaling
    # if adjust_B:
    #     start_B = na
    #     for k in range(udim):
    #         end_Bk = start_B + np.sum(nb[:k])
    #         THETA[start_B:end_Bk] *= y_std / U_std[k]
    NUM, DEN, NUMH, DENH = build_tfs(
        THETA, na, nb, nc, nd, nf, theta, id_method, udim, y_std, U_std
    )

    Vn = np.linalg.norm(y_id - y, 2) ** 2 / (2 * y.size)
    return (
        NUM.squeeze(),
        DEN.squeeze(),
        NUMH.squeeze(),
        DENH.squeeze(),
        Vn,
        y_id,
    )


def GEN_MISO_id(
    y: np.ndarray,
    u: np.ndarray,
    id_method: OptMethods,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: np.ndarray,
    max_iter: int,
    stab_marg: float,
    stab_cons: bool,
) -> tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.floating, np.ndarray
]:
    # Rescale inputs
    y_std, y = rescale(y)
    U_std = np.zeros(u.shape[0])
    for j in range(u.shape[0]):
        U_std[j], u[j] = rescale(u[j])

    check_valid_orders(u.shape[0], nb, theta)

    return GEN_id(
        y,
        u,
        id_method,
        na,
        nb,
        nc,
        nd,
        nf,
        theta,
        max_iter,
        stab_marg,
        stab_cons,
        False,
        y_std,
        U_std,
    )



"""
Created on Sat Nov 04 2017

@author: Giuseppe Armenise
"""

from warnings import warn

import numpy as np
import scipy as sc

from .functionset import information_criterion, rescale
from .functionsetSIM import (
    SS_lsim_predictor_form,
    SS_lsim_process_form,
    Vn_mat,
    Z_dot_PIort,
    impile,
    ordinate_sequence,
    reducingOrder,
)
from .typing import ICMethods, PARSIMMethods


def recalc_K(A, C, D, u):
    y_sim = []
    n_ord = A[:, 0].size
    m_input, L = u.shape
    l_ = C[:, 0].size
    n_simulations = n_ord + n_ord * m_input
    vect = np.zeros((n_simulations, 1))
    for i in range(n_simulations):
        vect[i, 0] = 1.0
        B = vect[0 : n_ord * m_input, :].reshape((n_ord, m_input))
        x0 = vect[n_ord * m_input : :, :].reshape((n_ord, 1))
        y_sim.append(
            (SS_lsim_process_form(A, B, C, D, u, x0=x0)[1]).reshape(
                (
                    1,
                    L * l_,
                )
            )
        )
        vect[i, 0] = 0.0
    y_matrix = 1.0 * y_sim[0]
    for j in range(n_simulations - 1):
        y_matrix = impile(y_matrix, y_sim[j + 1])
    y_matrix = y_matrix.T
    return y_matrix


def estimating_y(H_K, Uf, G_K, Yf, i, m, l_):
    y_tilde = np.dot(H_K[0:l_, :], Uf[m * (i) : m * (i + 1), :])
    for j in range(1, i):
        y_tilde = (
            y_tilde
            + np.dot(
                H_K[l_ * j : l_ * (j + 1), :],
                Uf[m * (i - j) : m * (i - j + 1), :],
            )
            + np.dot(
                G_K[l_ * j : l_ * (j + 1), :],
                Yf[l_ * (i - j) : l_ * (i - j + 1), :],
            )
        )
    return y_tilde


def estimating_y_S(H_K, Uf, Yf, i, m, l_):
    y_tilde = np.dot(H_K[0:l_, :], Uf[m * (i) : m * (i + 1), :])
    for j in range(1, i):
        y_tilde = y_tilde + np.dot(
            H_K[l_ * j : l_ * (j + 1), :], Uf[m * (i - j) : m * (i - j + 1), :]
        )
    return y_tilde


def SVD_weighted_K(Uf, Zp, Gamma_L):
    W2 = sc.linalg.sqrtm(np.dot(Z_dot_PIort(Zp, Uf), Zp.T)).real
    U_n, S_n, V_n = np.linalg.svd(np.dot(Gamma_L, W2), full_matrices=False)
    return U_n, S_n, V_n


def simulations_sequence(A_K, C, L, y, u, l_, m_, n, D_required):
    y_sim = []
    if D_required:
        n_simulations = n * m_ + l_ * m_ + n * l_ + n
        vect = np.zeros((n_simulations, 1))
        for i in range(n_simulations):
            vect[i, 0] = 1.0
            B_K = vect[0 : n * m_, :].reshape((n, m_))
            D = vect[n * m_ : n * m_ + l_ * m_, :].reshape((l_, m_))
            K = vect[n * m_ + l_ * m_ : n * m_ + l_ * m_ + n * l_, :].reshape(
                (
                    n,
                    l_,
                )
            )
            x0 = vect[n * m_ + l_ * m_ + n * l_ : :, :].reshape((n, 1))
            y_sim.append(
                (
                    SS_lsim_predictor_form(A_K, B_K, C, D, K, y, u, x0)[1]
                ).reshape((1, L * l_))
            )
            vect[i, 0] = 0.0
    else:
        D = np.zeros((l_, m_))
        n_simulations = n * m_ + n * l_ + n
        vect = np.zeros((n_simulations, 1))
        for i in range(n_simulations):
            vect[i, 0] = 1.0
            B_K = vect[0 : n * m_, :].reshape((n, m_))
            K = vect[n * m_ : n * m_ + n * l_, :].reshape((n, l_))
            x0 = vect[n * m_ + n * l_ : :, :].reshape((n, 1))
            y_sim.append(
                (
                    SS_lsim_predictor_form(A_K, B_K, C, D, K, y, u, x0)[1]
                ).reshape((1, L * l_))
            )
            vect[i, 0] = 0.0
    y_matrix = 1.0 * y_sim[0]
    for j in range(n_simulations - 1):
        y_matrix = impile(y_matrix, y_sim[j + 1])
    y_matrix = y_matrix.T
    return y_matrix


def simulations_sequence_S(A_K, C, L, K, y, u, l_, m_, n, D_required):
    y_sim = []
    if D_required:
        n_simulations = n * m_ + l_ * m_ + n
        vect = np.zeros((n_simulations, 1))
        for i in range(n_simulations):
            vect[i, 0] = 1.0
            B_K = vect[0 : n * m_, :].reshape((n, m_))
            D = vect[n * m_ : n * m_ + l_ * m_, :].reshape((l_, m_))
            x0 = vect[n * m_ + l_ * m_ : :, :].reshape((n, 1))
            y_sim.append(
                (
                    SS_lsim_predictor_form(A_K, B_K, C, D, K, y, u, x0)[1]
                ).reshape((1, L * l_))
            )
            vect[i, 0] = 0.0
    else:
        n_simulations = n * m_ + n
        vect = np.zeros((n_simulations, 1))
        D = np.zeros((l_, m_))
        for i in range(n_simulations):
            vect[i, 0] = 1.0
            B_K = vect[0 : n * m_, :].reshape((n, m_))
            x0 = vect[n * m_ : :, :].reshape((n, 1))
            y_sim.append(
                (
                    SS_lsim_predictor_form(A_K, B_K, C, D, K, y, u, x0)[1]
                ).reshape((1, L * l_))
            )
            vect[i, 0] = 0.0
    y_matrix = 1.0 * y_sim[0]
    for j in range(n_simulations - 1):
        y_matrix = impile(y_matrix, y_sim[j + 1])
    y_matrix = y_matrix.T
    return y_matrix


def AK_C_estimating_S_P(U_n, S_n, V_n, l_, f, m, Zp, Uf, Yf):
    n = S_n.size
    S_n = np.diag(S_n)
    Ob_f = np.dot(U_n, sc.linalg.sqrtm(S_n))
    A = np.dot(np.linalg.pinv(Ob_f[0 : l_ * (f - 1), :]), Ob_f[l_::, :])
    C = Ob_f[0:l_, :]
    Q, R = np.linalg.qr(impile(impile(Zp, Uf), Yf).T)
    Q = Q.T
    R = R.T
    G_f = R[(2 * m + l_) * f : :, (2 * m + l_) * f : :]
    F = G_f[0:l_, 0:l_]
    K = np.dot(
        np.dot(np.linalg.pinv(Ob_f[0 : l_ * (f - 1), :]), G_f[l_::, 0:l_]),
        np.linalg.inv(F),
    )
    A_K = A - np.dot(K, C)
    return A, C, A_K, K, n


def sim_observed_seq(y, u, f, D_required, l_, L, m, U_n, S_n):
    n = S_n.size
    S_n = np.diag(S_n)
    Ob_K = np.dot(U_n, sc.linalg.sqrtm(S_n))
    A_K = np.dot(np.linalg.pinv(Ob_K[0 : l_ * (f - 1), :]), Ob_K[l_::, :])
    C = Ob_K[0:l_, :]
    y_sim = simulations_sequence(A_K, C, L, y, u, l_, m, n, D_required)
    return y_sim, A_K, C


def parsim(
    y: np.ndarray,
    u: np.ndarray,
    mode: PARSIMMethods,
    order: int | tuple[int, int] = 0,
    threshold: float = 0.0,
    f: int = 20,
    p: int = 20,
    D_required: bool = False,
    B_recalc: bool = False,
    ic_method: ICMethods = "AIC",
):
    if isinstance(order, tuple):
        min_ord, max_ord = order[0], order[-1] + 1
        if min_ord < 1:
            warn("The minimum model order will be set to 1")
            min_ord = 1
        if f < min_ord:
            warn(
                f"The horizon must be larger than the model order, min_order set to {f}"
            )
            min_ord = f
        if f < max_ord - 1:
            warn(
                f"The horizon must be larger than the model order, max_order set to {f}"
            )
            max_ord = f + 1

    y = np.atleast_2d(y)
    u = np.atleast_2d(u)

    l_, L = y.shape
    m_ = u[:, 0].size
    U_std = np.zeros(m_)
    Ystd = np.zeros(l_)
    for j in range(m_):
        U_std[j], u[j] = rescale(u[j])
    for j in range(l_):
        Ystd[j], y[j] = rescale(y[j])
    Yf, Yp = ordinate_sequence(y, f, p)
    Uf, Up = ordinate_sequence(u, f, p)
    Zp = impile(Up, Yp)

    Gamma_L = compute_gamma_matrix(mode, f, l_, m_, Yf, Uf, Zp)

    U_n, S_n, V_n = SVD_weighted_K(Uf, Zp, Gamma_L)

    if isinstance(order, tuple):
        min_order = order[0]
        IC_old = np.inf
        for i in range(min_ord, max_ord):
            U_n, S_n, V_n = reducingOrder(U_n, S_n, V_n, threshold, i)
            if mode == "PARSIM_K":
                y_sim, A_K, C = sim_observed_seq(
                    y, u, f, D_required, l_, L, m_, U_n, S_n
                )

            else:
                A, C, A_K, K, n = AK_C_estimating_S_P(
                    U_n, S_n, V_n, l_, f, m_, Zp, Uf, Yf
                )
                y_sim = simulations_sequence_S(
                    A_K, C, L, K, y, u, l_, m_, n, D_required
                )
            vect = np.dot(np.linalg.pinv(y_sim), y.reshape((L * l_, 1)))
            Y_estimate = np.dot(y_sim, vect)
            Vn = Vn_mat(y.reshape((L * l_, 1)), Y_estimate)

            K_par = 2 * n * l_ + m_ * n
            if D_required:
                K_par = K_par + l_ * m_
            IC = information_criterion(K_par, L, Vn, ic_method)
            if IC < IC_old:
                min_order = i
                IC_old = IC

        order = min_order
        print("The suggested order is: n=", order)

    U_n, S_n, V_n = reducingOrder(U_n, S_n, V_n, threshold, order)
    n = S_n.size
    if mode == "PARSIM_K":
        y_sim, A_K, C = sim_observed_seq(
            y, u, f, D_required, l_, L, m_, U_n, S_n
        )

    else:
        A, C, A_K, K, n = AK_C_estimating_S_P(
            U_n, S_n, V_n, l_, f, m_, Zp, Uf, Yf
        )
        y_sim = simulations_sequence_S(
            A_K, C, L, K, y, u, l_, m_, n, D_required
        )

    vect = np.dot(np.linalg.pinv(y_sim), y.reshape((L * l_, 1)))
    Y_estimate = np.dot(y_sim, vect)
    Vn = Vn_mat(y.reshape((L * l_, 1)), Y_estimate)
    B_K = vect[0 : n * m_, :].reshape((n, m_))

    if D_required:
        D = vect[n * m_ : n * m_ + l_ * m_, :].reshape((l_, m_))
        if mode == "PARSIM_K":
            K = vect[n * m_ + l_ * m_ : n * m_ + l_ * m_ + n * l_, :].reshape(
                (
                    n,
                    l_,
                )
            )
            x0 = vect[n * m_ + l_ * m_ + n * l_ : :, :].reshape((n, 1))
        else:
            x0 = vect[n * m_ + l_ * m_ : :, :].reshape((n, 1))
    else:
        D = np.zeros((l_, m_))
        if mode == "PARSIM_K":
            K = vect[n * m_ : n * m_ + n * l_, :].reshape((n, l_))
            x0 = vect[n * m_ + n * l_ : :, :].reshape((n, 1))
        else:
            x0 = vect[n * m_ : :, :].reshape((n, 1))

    if mode == "PARSIM_K":
        A = A_K + np.dot(K, C)
        if B_recalc:
            y_sim = recalc_K(A, C, D, u)
            vect = np.dot(np.linalg.pinv(y_sim), y.reshape((L * l_, 1)))
            Y_estimate = np.dot(y_sim, vect)
            Vn = Vn_mat(y.reshape((L * l_, 1)), Y_estimate)
            B = vect[0 : n * m_, :].reshape((n, m_))
            x0 = vect[n * m_ : :, :].reshape((n, 1))
            B_K = B - np.dot(K, D)

    for j in range(m_):
        B_K[:, j] = B_K[:, j] / U_std[j]
        D[:, j] = D[:, j] / U_std[j]
    for j in range(l_):
        K[:, j] = K[:, j] / Ystd[j]
        C[j, :] = C[j, :] * Ystd[j]
        D[j, :] = D[j, :] * Ystd[j]
    B = B_K + np.dot(K, D)
    return A_K, C, B_K, D, K, A, B, x0, Vn


def compute_gamma_matrix(mode, f, l_, m_, Yf, Uf, Zp):
    if mode == "PARSIM_K":
        M = np.dot(Yf[0:l_, :], np.linalg.pinv(impile(Zp, Uf[0:m_, :])))
        Matrix_pinv = np.linalg.pinv(
            impile(Zp, impile(Uf[0:m_, :], Yf[0:l_, :]))
        )
    else:
        Matrix_pinv = np.linalg.pinv(impile(Zp, Uf[0:m_, :]))
        M = np.dot(Yf[0:l_, :], Matrix_pinv)
    Gamma_L = M[:, 0 : (m_ + l_) * f]

    H = M[:, (m_ + l_) * f : :]
    G = np.zeros((l_, l_))
    for i in range(1, f):
        if mode == "PARSIM_K":
            y_tilde = estimating_y(H, Uf, G, Yf, i, m_, l_)
            M = np.dot((Yf[l_ * i : l_ * (i + 1)] - y_tilde), Matrix_pinv)
            H = impile(H, M[:, (m_ + l_) * f : (m_ + l_) * f + m_])
            G = impile(G, M[:, (m_ + l_) * f + m_ : :])
        elif mode == "PARSIM_S":
            y_tilde = estimating_y_S(H, Uf, Yf, i, m_, l_)
            M = np.dot((Yf[l_ * i : l_ * (i + 1)] - y_tilde), Matrix_pinv)
            H = impile(H, M[:, (m_ + l_) * f : :])
        elif mode == "PARSIM_P":
            Matrix_pinv = np.linalg.pinv(impile(Zp, Uf[0 : m_ * (i + 1), :]))
            M = np.dot((Yf[l_ * i : l_ * (i + 1)]), Matrix_pinv)
        Gamma_L = impile(Gamma_L, (M[:, 0 : (m_ + l_) * f]))
    return Gamma_L



from warnings import warn

import control as cnt
import numpy as np


def atleast_3d(arr: list | np.ndarray) -> np.ndarray:
    arr = np.array(arr)
    if arr.ndim == 1:
        return arr.reshape(1, 1, -1)
    elif arr.ndim == 2:
        return arr.reshape(1, *arr.shape)
    else:
        return arr


def check_valid_orders(dim: int, *orders: np.ndarray):
    for i, arg in enumerate(orders):
        if isinstance(arg, int) or arg.shape == ():
            continue

        if arg.shape[0] != dim:
            arg_is_vec = len(arg.shape) == 1
            raise RuntimeError(
                f"Argument {i} must be a {'vector' if arg_is_vec else 'matrix'}, whose dimensions must be equal to {dim}"
            )
        if not np.issubdtype(arg.dtype, np.integer) or np.min(arg) < 0:
            raise RuntimeError(
                f"Arguments must contain only positive int elements. Arg {i} violates this rule."
            )


def check_feasibility(G, H, id_method: str, stab_marg: float, stab_cons: bool):
    poles_G = np.abs(cnt.poles(G))
    poles_H = np.abs(cnt.poles(H))

    if len(poles_G) != 0 and len(poles_H) != 0:
        poles_G = max(poles_G)
        poles_H = max(poles_H)
        # TODO: verify with RBdC if correct setting this to zero. Raises warnings.
        # check_st_H = poles_H
        if poles_G > 1.0 or poles_H > 1.0:
            warn("One of the identified system is not stable")
            if stab_cons is True:
                raise RuntimeError(
                    f"Infeasible solution: the stability constraint has been violated, since the maximum pole is {max(poles_H, poles_G)} \
                        ... against the imposed stability margin {stab_marg}"
                )
            else:
                warn(
                    f"Consider activating the stability constraint. The maximum pole is {max(poles_H, poles_G)}  "
                )


def get_val_range(order_range: int | tuple[int, int]):
    if isinstance(order_range, int):
        order_range = (order_range, order_range + 1)
    min_val, max_val = order_range
    if min_val < 0:
        raise ValueError("Minimum value must be non-negative")
    return range(min_val, max_val + 1)


def validate_and_prepare_inputs(
    u: np.ndarray, nb: int | np.ndarray, theta: int | np.ndarray
) -> tuple[np.ndarray, np.ndarray, np.ndarray, int]:
    """Check input dimensions and ensure nb/theta are arrays."""
    u = np.atleast_2d(u)
    udim = u.shape[0]
    nb = np.atleast_1d(nb)
    theta = np.atleast_1d(theta)
    check_valid_orders(udim, nb, theta)
    return u, nb, theta, udim


def common_setup(
    na: int,
    nb: int | np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: int | np.ndarray,
) -> tuple[int, int]:
    nbth = nb + theta
    val = max(na, np.max(nbth), nc, nd, nf)
    n_coeff = na + np.sum(nb) + nc + nd + nf
    return val, n_coeff


def build_polynomial(order: int, coeffs: np.ndarray) -> cnt.TransferFunction:
    """
    Build a polynomial transfer function.
    This function constructs a transfer function of the form:
    H(s) = (s^order + 0*s^(order-1) + ... + 0*s + 1) / (s^order + coeffs[0]*s^(order-1) + ... + coeffs[order-1])
    Args:
        order (int): The order of the polynomial.
        coeffs (np.ndarray): The coefficients of the polynomial.
    Returns:
        cnt.TransferFunction: The resulting transfer function.
    Raises:
        RuntimeError: If the transfer function could not be obtained.
    Example:
        >>> import numpy as np
        >>> import control as cnt
        >>> coeffs = np.array([1, 2, 3])
        >>> tf = build_polynomial(3, coeffs)
        >>> tf
        TransferFunction(array([1, 0, 0, 0]), array([1, 1, 2, 3]), 1)
    """

    tf = cnt.tf([1] + [0] * order, [1] + list(coeffs), 1)
    if tf is None:
        raise RuntimeError("TF could not be obtained")
    return tf


def build_tf_G(
    THETA: np.ndarray,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: np.ndarray,
    id_method: str,
    udim: int,
    y_std: float = 1.0,
    U_std: np.ndarray = np.array([1.0]),
) -> tuple[np.ndarray, np.ndarray]:
    """Construct NUM, DEN, NUMH, DENH from parameters."""
    ng = na if id_method != "OE" else nf
    nb_total = np.sum(nb)
    nf_start = 0 if id_method == "OE" else na + nb_total + nc + nd
    indices = {
        "A": (0, na),
        "B": (ng, ng + nb_total),
        "F": (nf_start, nf_start + nf),
    }

    # Denominator polynomials
    A = build_polynomial(na, THETA[indices["A"][0] : indices["A"][1]])
    F = build_polynomial(nf, THETA[indices["F"][0] : indices["F"][1]])

    # Denominator calculations
    denG = np.array(cnt.tfdata(A * F)[1][0]) if A and F else np.array([1])

    # Numerator handling
    valG = max(np.max(nb + theta), na + nf)

    if id_method == "ARMA":
        NUM = np.ones((udim, 1))
    else:
        NUM = np.zeros((udim, valG))

    DEN = np.zeros((udim, valG + 1))

    for k in range(udim):
        if id_method != "ARMA":
            # TODO: verify whether this adjustment should be done prior to using THETA for polynomial calculations
            #  actual implementation is consistent with version 0.*.* of SIPPY
            b_slice = (
                THETA[
                    indices["B"][0] + np.sum(nb[:k]) : indices["B"][0]
                    + np.sum(nb[: k + 1])
                ]
                * y_std
                / U_std[k]
            )
            NUM[k, theta[k] : theta[k] + nb[k]] = b_slice

        DEN[k, : na + nf + 1] = denG

    return NUM, DEN


def build_tf_H(
    THETA: np.ndarray,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    _: int,
    __: np.ndarray,
    id_method: str,
    ___: int,
) -> tuple[np.ndarray, np.ndarray]:
    """Construct NUM, DEN, NUMH, DENH from parameters."""
    nb_total = np.sum(nb)
    indices = {
        "A": (0, na),
        "C": (na + nb_total, na + nb_total + nc),
        "D": (na + nb_total + nc, na + nb_total + nc + nd),
    }

    # Denominator polynomials
    A = build_polynomial(na, THETA[indices["A"][0] : indices["A"][1]])
    D = build_polynomial(nd, THETA[indices["D"][0] : indices["D"][1]])

    # Denominator calculations
    denH = np.array(cnt.tfdata(A * D)[1][0]) if A and D else np.array([1])

    # Numerator handling
    valH = max(nc, na + nd)

    if id_method == "OE":
        NUMH = np.ones((1, 1))
    else:
        NUMH = np.zeros((1, valH + 1))
        NUMH[0, 0] = 1.0
        NUMH[0, 1 : nc + 1] = THETA[indices["C"][0] : indices["C"][1]]

    DENH = np.zeros((1, valH + 1))
    DENH[0, : na + nd + 1] = denH

    return NUMH, DENH


def build_tfs(
    THETA: np.ndarray,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: np.ndarray,
    id_method: str,
    udim: int,
    y_std: float = 1.0,
    U_std: np.ndarray = np.array([1.0]),
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Construct NUM, DEN, NUMH, DENH from parameters."""
    NUM, DEN = build_tf_G(
        THETA, na, nb, nc, nd, nf, theta, id_method, udim, y_std, U_std
    )
    NUMH, DENH = build_tf_H(THETA, na, nb, nc, nd, nf, theta, id_method, udim)

    return NUM, DEN, NUMH, DENH



"""
Created on 2021

@author: RBdC & MV
"""

import numpy as np
from casadi import DM, SX, Function, mtimes, nlpsol, norm_inf, vertcat

from .typing import OptMethods


# Defining the optimization problem
def opt_id(
    Y: np.ndarray,
    U: np.ndarray,
    FLAG: OptMethods,
    na: int,
    nb: np.ndarray,
    nc: int,
    nd: int,
    nf: int,
    theta: np.ndarray,
    max_iter: int,
    stab_marg: float,
    stab_cons: bool,
    n_coeff,
    m,
    n_tr,
) -> tuple[Function, DM, DM, DM, DM]:
    # orders
    nb_ = np.sum(nb)

    # Augment the optmization variables with Y vector to build a multiple shooting problem
    N = Y.size

    # Augment the optmization variables with auxiliary variables
    if nd != 0:
        n_aus = 3 * N
    else:
        n_aus = N

    # Optmization variables
    n_opt = n_aus + n_coeff

    # Define symbolic optimization variables
    w_opt = SX.sym("w", n_opt)

    # Build optimization variable
    # Get subset a
    a = w_opt[0:na]

    # Get subset b
    b = w_opt[na : na + nb_]

    # Get subsets c and d
    c = w_opt[na + nb_ : na + nb_ + nc]
    d = w_opt[na + nb_ + nc : na + nb_ + nc + nd]

    # Get subset f
    f = w_opt[na + nb_ + nd + nc : na + nb_ + nc + nd + nf]

    # Optimization variables
    Yidw = w_opt[-N:]

    # Additional optimization variables
    if nd != 0:
        Ww = w_opt[-3 * N : -2 * N]
        Vw = w_opt[-2 * N : -N]

    # Initializing bounds on optimization variables
    w_lb = -1e0 * DM.inf(n_opt)
    w_ub = 1e0 * DM.inf(n_opt)
    #
    w_lb = -1e2 * DM.ones(n_opt)
    w_ub = 1e2 * DM.ones(n_opt)

    # Build Regressor
    # depending on the model structure

    # Building coefficient vector
    if FLAG == "OE":
        coeff = vertcat(b, f)
    elif FLAG == "BJ":
        coeff = vertcat(b, f, c, d)
    elif FLAG == "ARMAX":
        coeff = vertcat(a, b, c)
    elif FLAG == "ARARX":
        coeff = vertcat(a, b, d)
    elif FLAG == "ARARMAX":
        coeff = vertcat(a, b, c, d)
    elif FLAG == "ARMA":
        coeff = vertcat(a, c)
    else:  # GEN
        coeff = vertcat(a, b, f, c, d)

    # Define Yid output model
    Yid = Y * SX.ones(1)

    # Preallocate internal variables
    if nd != 0:
        W = Y * SX.ones(1)  # w = B * u or w = B/F * u
        V = Y * SX.ones(1)  # v = A*y - w

        if na != 0:
            coeff_v = a
        if nf != 0:  # BJ, GEN
            coeff_w = vertcat(b, f)
        else:  # ARARX, ARARMAX
            coeff_w = vertcat(b)

    if nc != 0:
        Epsi = SX.zeros(N)

    for k in range(N):
        # n_tr: number of not identifiable outputs
        if k >= n_tr:
            # building regressor
            if nb_ != 0:
                # inputs
                vecU = []
                for nb_i in range(m):
                    vecu = U[nb_i, :][
                        k - nb[nb_i] - theta[nb_i] : k - theta[nb_i]
                    ][::-1]
                    vecU = vertcat(vecU, vecu)

            # measured output Y
            if na != 0:
                vecY = Y[k - na : k][::-1]

            # auxiliary variable V
            if nd != 0:
                vecV = Vw[k - nd : k][::-1]

                # auxiliary variable W
                if nf != 0:
                    vecW = Ww[k - nf : k][::-1]

            # prediction error
            if nc != 0:
                vecE = Epsi[k - nc : k][::-1]

            # regressor
            if FLAG == "OE":
                vecY = Yidw[k - nf : k][::-1]
                phi = vertcat(vecU, -vecY)
            elif FLAG == "BJ":
                phi = vertcat(vecU, -vecW, vecE, -vecV)
            elif FLAG == "ARMAX":
                phi = vertcat(-vecY, vecU, vecE)
            elif FLAG == "ARMA":
                phi = vertcat(-vecY, vecE)
            elif FLAG == "ARARX":
                phi = vertcat(-vecY, vecU, -vecV)
            elif FLAG == "ARARMAX":
                phi = vertcat(-vecY, vecU, vecE, -vecV)
            else:
                phi = vertcat(-vecY, vecU, -vecW, vecE, -vecV)

            # update prediction
            Yid[k] = mtimes(phi.T, coeff)

            # pred. error
            if nc != 0:
                Epsi[k] = Y[k] - Yidw[k]

            # auxiliary variable W
            if nd != 0:
                if nf != 0:
                    phiw = vertcat(vecU, -vecW)  # BJ, GEN
                else:
                    phiw = vertcat(vecU)  # ARARX, ARARMAX
                W[k] = mtimes(phiw.T, coeff_w)

                # auxiliary variable V
                if na == 0:  # 'BJ'  [A(z) = 1]
                    V[k] = Y[k] - Ww[k]
                else:  # [A(z) div 1]
                    phiv = vertcat(vecY)
                    V[k] = Y[k] + mtimes(phiv.T, coeff_v) - Ww[k]

    # Objective Function
    DY = Y - Yidw

    f_obj = (1.0 / (N)) * mtimes(DY.T, DY)

    # if  FLAG != 'ARARX' or FLAG != 'OE':
    #   f_obj += 1e-4*mtimes(c.T,c)   # weighting c

    # Getting constrains
    g = []

    # Equality constraints
    g.append(Yid - Yidw)

    if nd != 0:
        g.append(W - Ww)
        g.append(V - Vw)

    # Stability check
    ng_norm = 0
    if stab_cons is True:
        if na != 0:
            ng_norm += 1
            # companion matrix A(z) polynomial
            compA = SX.zeros(na, na)
            diagA = SX.eye(na - 1)
            compA[:-1, 1:] = diagA
            compA[-1, :] = -a[::-1]  # opposite reverse coeficient a

            # infinite-norm
            norm_CompA = norm_inf(compA)

            # append on eq. constraints
            g.append(norm_CompA)

        if nf != 0:
            ng_norm += 1
            # companion matrix F(z) polynomial
            compF = SX.zeros(nf, nf)
            diagF = SX.eye(nf - 1)
            compF[:-1, 1:] = diagF
            compF[-1, :] = -f[::-1]  # opposite reverse coeficient f

            # infinite-norm
            norm_CompF = norm_inf(compF)

            # append on eq. constraints
            g.append(norm_CompF)

        if nd != 0:
            ng_norm += 1
            # companion matrix D(z) polynomial
            compD = SX.zeros(nd, nd)
            diagD = SX.eye(nd - 1)
            compD[:-1, 1:] = diagD
            compD[-1, :] = -d[::-1]  # opposite reverse coeficient D

            # infinite-norm
            norm_CompD = norm_inf(compD)

            # append on eq. constraints
            g.append(norm_CompD)

    # constraint vector
    g = vertcat(*g)

    # Constraint bounds
    ng = g.size1()
    g_lb = -1e-7 * DM.ones(ng, 1)
    g_ub = 1e-7 * DM.ones(ng, 1)

    # Force system stability
    # note: norm_inf(X) >= Spectral radius (A)
    if ng_norm != 0:
        g_ub[-ng_norm:] = stab_marg * DM.ones(ng_norm, 1)
        # for i in range(ng_norm):
        #     f_obj += 1e1*fmax(0,g_ub[-i-1:]-g[-i-1:])

    # NL optimization variables
    nlp = {"x": w_opt, "f": f_obj, "g": g}

    # Solver options
    # sol_opts = {'ipopt.max_iter':max_iter}#, 'ipopt.tol':1e-10}#,'ipopt.print_level':0,'ipopt.sb':"yes",'print_time':0}
    sol_opts = {
        "ipopt.max_iter": max_iter,
        "ipopt.print_level": 0,
        "ipopt.sb": "yes",
        "print_time": 0,
    }

    # Defining the solver
    solver = nlpsol("solver", "ipopt", nlp, sol_opts)

    return solver, w_lb, w_ub, g_lb, g_ub



from typing import Literal, TypedDict

IOMethods = Literal[
    "FIR",
    "ARX",
    "ARMA",
    "ARMAX",
    "OE",
    "ARARX",
    "ARARMAX",
    "BJ",
    "GEN",
    "EARMAX",
    "EOE",
]

OLSimMethods = Literal["CVA", "MOESP", "N4SID"]
PARSIMMethods = Literal["PARSIM_K", "PARSIM_P", "PARSIM_S"]
# TODO: Use | if issue resolved https://github.com/python/typing/issues/1932
SSMethods = Literal[
    "CVA", "MOESP", "N4SID", "PARSIM_K", "PARSIM_P", "PARSIM_S"
]
# TODO: Use | if issue resolved https://github.com/python/typing/issues/1932
AvailableMethods = Literal[
    "FIR",
    "ARX",
    "ARMA",
    "ARMAX",
    "OE",
    "ARARX",
    "ARARMAX",
    "BJ",
    "GEN",
    "EARMAX",
    "EOE",
    "CVA",
    "MOESP",
    "N4SID",
    "PARSIM_K",
    "PARSIM_P",
    "PARSIM_S",
]

ArxMethods = Literal["FIR", "ARX", "ARMA", "ARARX", "ARARMAX"]
OptMethods = Literal[
    "OE", "BJ", "GEN", "ARARX", "ARARMAX", "ARMA", "ARMAX", "EARMAX", "EOE"
]

RLSMethods = Literal["ARX", "ARMA", "ARMAX", "ARARX", "ARARMAX", "FIR", "OE"]

ICMethods = Literal["AIC", "AICc", "BIC"]

AvailableModes = Literal["LLS", "ILLS", "RLLS", "OPT"]

Flags = Literal["arx", "armax", "rls", "opt"]

CenteringMethods = Literal["InitVal", "MeanVal", None]


class OrderRanges(TypedDict, total=False):
    na_ord: tuple[int, int]
    nb_ord: tuple[int, int]
    nc_ord: tuple[int, int]
    nd_ord: tuple[int, int]
    nf_ord: tuple[int, int]
    delays: tuple[int, int]


class FlexOrderParams(TypedDict, total=False):
    IC: ICMethods
    id_mode: AvailableModes
    centering: CenteringMethods



"""
Created on Fri Jul 28 2017

@author: Giuseppe Armenise
"""

from warnings import warn

import control.matlab as cnt
import numpy as np

from .functionset import mean_square_error, rescale
from .utils import get_val_range


def ARMAX_MISO_id(
    y: np.ndarray,
    u: np.ndarray,
    na: int,
    nb: np.ndarray,
    nc: int,
    theta: np.ndarray,
    max_iter: int,
):
    nb = np.array(nb)
    theta = np.array(theta)
    u = np.atleast_2d(u)
    ylength = y.size
    y_std, y = rescale(y)
    [udim, ulength] = u.shape
    eps = np.zeros(y.size)
    Reached_max = False
    # checking dimension
    if nb.size != udim:
        raise RuntimeError(
            " nb must be a matrix, whose dimensions must be equal to yxu"
        )
    #        return np.array([[1.]]),np.array([[0.]]),np.array([[0.]]),np.inf,Reached_max
    elif theta.size != udim:
        raise RuntimeError("theta matrix must have yxu dimensions")
    #        return np.array([[1.]]),np.array([[0.]]),np.array([[0.]]),np.inf,Reached_max
    else:
        nbth = nb + theta
        U_std = np.zeros(udim)
        for j in range(udim):
            U_std[j], u[j] = rescale(u[j])
        # TODO: if run as MIMO, this should be fixed
        val = max(na, np.max(nbth), nc)
        # max predictable dimension
        N = ylength - val
        # regressor matrix
        phi = np.zeros(na + np.sum(nb[:]) + nc)
        PHI = np.zeros((N, na + np.sum(nb[:]) + nc))
        for k in range(N):
            phi[0:na] = -y[k + val - 1 :: -1][0:na]
            for nb_i in range(udim):
                phi[
                    na + np.sum(nb[0:nb_i]) : na + np.sum(nb[0 : nb_i + 1])
                ] = u[nb_i, :][val + k - 1 :: -1][
                    theta[nb_i] : nb[nb_i] + theta[nb_i]
                ]
            PHI[k, :] = phi
        Vn = np.inf
        Vn_old = np.inf
        # coefficient vector
        THETA = np.zeros(na + np.sum(nb[:]) + nc)
        ID_THETA = np.identity(THETA.size)
        lambdak = 0.5
        iterations = 0
        while (Vn_old > Vn or iterations == 0) and iterations < max_iter:
            THETA_old = THETA
            Vn_old = Vn
            iterations = iterations + 1
            for i in range(N):
                PHI[i, na + np.sum(nb[:]) : na + np.sum(nb[:]) + nc] = eps[
                    val + i - 1 :: -1
                ][0:nc]
            THETA = np.dot(np.linalg.pinv(PHI), y[val::])
            Vn = (np.linalg.norm(y[val::] - np.dot(PHI, THETA), 2) ** 2) / (
                2 * N
            )
            THETA_new = THETA
            lambdak = 0.5
            while Vn > Vn_old:
                THETA = np.dot(ID_THETA * lambdak, THETA_new) + np.dot(
                    ID_THETA * (1 - lambdak), THETA_old
                )
                # Model Output
                # y_id0 = np.dot(PHI,THETA)
                Vn = (
                    np.linalg.norm(y[val::] - np.dot(PHI, THETA), 2) ** 2
                ) / (2 * N)

                if lambdak < np.finfo(np.float32).eps:
                    THETA = THETA_old
                    Vn = Vn_old
                lambdak = lambdak / 2.0
            eps[val::] = y[val::] - np.dot(PHI, THETA)
            # adding non-identified outputs
            y_id = np.hstack((y[:val], np.dot(PHI, THETA))) * y_std

        if iterations >= max_iter:
            warn("Reached maximum iterations")
            Reached_max = True
        denominator = np.zeros((udim, val + 1))
        numerator_H = np.zeros((1, val + 1))
        numerator_H[0, 0] = 1.0
        numerator_H[0, 1 : nc + 1] = THETA[na + np.sum(nb[:]) : :]
        denominator[:, 0] = np.ones(udim)
        numerator = np.zeros((udim, val))
        for k in range(udim):
            THETA[na + np.sum(nb[0:k]) : na + np.sum(nb[0 : k + 1])] = (
                THETA[na + np.sum(nb[0:k]) : na + np.sum(nb[0 : k + 1])]
                * y_std
                / U_std[k]
            )
            numerator[k, theta[k] : theta[k] + nb[k]] = THETA[
                na + np.sum(nb[0:k]) : na + np.sum(nb[0 : k + 1])
            ]
            denominator[k, 1 : na + 1] = THETA[0:na]
        denominator_H = denominator[0]
        return (
            numerator,
            denominator,
            numerator_H,
            denominator_H,
            Vn,
            y_id,
            Reached_max,
        )


class Armax:
    def __init__(
        self,
        G: cnt.TransferFunction,
        H: cnt.TransferFunction,
        *order_bounds: tuple[int, int],
        Vn,
        Yid,
        method="AIC",
    ):
        """Armax model class.

        The AutoRegressive-Moving-Average with eXogenous inputs model is computed based on a
        recursive lest-square regression between the input data (U) and the measured output data
        (Y). As Y is noisy in practice, a white noise (E) is identified within the model.
        This model is designed to deal with potential thetas between U and Y.

        The following equations summarize the equations involved in the model:

        Y = G.U + H.E

        G = B / A
        H = C / A

        A = 1 + a_1*z^(-1) + ... + a_na*z^(-na)
        B = b_1*z^(-1-theta) + ... + b_nb*z^(-nb-theta)
        C = c_1*z^(-1) + ... + c_nc*z^(-nc)

        .. seealso:: https://ieeexplore.ieee.org/abstract/document/8516791


        :param na_bounds: extended range of the order of the common denominator
        :type na_bounds: list of two ints
        :param nb_bounds: extended range of the order of the G numerator
        :type nb_bounds: list of two ints
        :param nc_bounds: extended range of the order of the H numerator
        :type nc_bounds: list of two ints
        :param theta_bounds: extended range of the discrete theta in B
        :type theta_bounds: list of two ints
        :param dt: sampling time of discretized data y and u
        :type dt: float
        :param method: Method used of to attribute a performance to the model
        :type method: string
        :param max_iter: maximum numbers of iterations to find the best fit
        :type max_iter: int
        """
        for order in order_bounds:
            if isinstance(order, list | tuple):
                if not all(isinstance(x, int) for x in order):
                    raise ValueError(
                        "wrong arguments passed to define an armax model"
                    )

        self.orders = [get_val_range(order) for order in order_bounds]

        self.method = method

        self.G = G
        self.H = H
        self.Vn = Vn
        self.Yid = Yid

    def __repr__(self):
        return f"Armax({', '.join(f'[{min(order)}, {max(order)}]' for order in self.orders)}, {self.method})"

    def __str__(self):
        return (
            "Armax model:\n"
            "- Params:\n"
            f"  orders: {', '.join(f'[{min(order)}, {max(order)}]' for order in self.orders)} \n"
            f"  dt: {self.G.dt} \n"
            f"  method: {self.method} \n"
            "- Output:\n"
            f"  G: {self.G} \n"
            f"  H: {self.H} \n"
            f"  Vn: {self.Vn} \n"
            f"  Model Output: {self.Yid} \n"
        )

    @staticmethod
    def _identify(
        y: np.ndarray,
        u: np.ndarray,
        na: int,
        nb: int | np.ndarray,
        nc: int,
        theta: int | np.ndarray,
        max_iter: int,
    ):
        """Identify

        Given model order as parameter, the recursive algorithm looks for the best fit in less
        than max_iter steps.

        At each step, the algorithm performs a least-square regression. If the

        :param y: Measured data
        :type y: Array of float
        :param u: Input data
        :type u: Array of float (same shape as y)
        :param na: order of the common denominator
        :type na: int
        :param nb: order of the numerator of G
        :type nb: int
        :param nc: order of the numerator of H
        :type nc: int
        :param theta: discrete theta expressed as a number of shifted indices between u and y
        :type theta: int
        :param max_iter: maximum numbers of iterations to find the best fit
        :type max_iter: int
        :return numerator: Numerator of G
        :rtype numerator: float array
        :return denominator: Denominator of G
        :rtype denominator: float array
        :return numerator_H: Numerator of H
        :rtype numerator_H: float array
        :return denominator_H: Denominator of H
        :rtype denominator_H: float array
        :return Vn: variance between y and the estimated output data (X*beta_hat)
        :rtype Vn: float
        :return Yid: estimated output data (X*beta_hat)
        :rtype Yid: float array
        :return max_reached: Whether the algorithm reached maximum number of iterations or not.
        :rtype max_reached: boolean
        """
        if isinstance(nb, np.ndarray):
            nb = int(np.sum(nb))
        if isinstance(theta, np.ndarray):
            theta = int(np.sum(theta))
        max_order = max(na, nb + theta, nc)
        sum_order = sum((na, nb, nc))

        # Define the usable measurements length, N, for the identification process
        N: int = y.size - max_order

        noise_hat = np.zeros(y.size)

        # Fill X matrix used to perform least-square regression: beta_hat = (X_T.X)^(-1).X_T.y
        X = np.zeros((N, sum_order))
        for i in range(N):
            X[i, 0:na] = -y[i + max_order - 1 :: -1][0:na]
            X[i, na : na + nb] = u[max_order + i - 1 :: -1][theta : nb + theta]

        Vn, Vn_old = np.inf, np.inf
        beta_hat = np.zeros(sum_order)
        I_beta = np.identity(beta_hat.size)
        iterations = 0
        # max_reached = False

        # Stay in this loop while variance has not converged or max iterations has not been
        # reached yet.
        while (Vn_old > Vn or iterations == 0) and iterations < max_iter:
            beta_hat_old = beta_hat
            Vn_old = Vn
            iterations = iterations + 1
            for i in range(N):
                X[i, na + nb : na + nb + nc] = noise_hat[
                    max_order + i - 1 :: -1
                ][0:nc]
            beta_hat = np.dot(np.linalg.pinv(X), y[max_order::])
            Vn = mean_square_error(y[max_order::], np.dot(X, beta_hat))

            # If solution found is not better than before, perform a binary search to find a better
            # solution.
            beta_hat_new = beta_hat
            interval_length = 0.5
            while Vn > Vn_old:
                beta_hat = np.dot(
                    I_beta * interval_length, beta_hat_new
                ) + np.dot(I_beta * (1 - interval_length), beta_hat_old)
                Vn = mean_square_error(y[max_order::], np.dot(X, beta_hat))

                # Stop the binary search when the interval length is minor than smallest float
                if interval_length < np.finfo(np.float32).eps:
                    beta_hat = beta_hat_old
                    Vn = Vn_old
                interval_length = interval_length / 2.0

            # Update estimated noise based on best solution found from currently considered
            # noise.
            noise_hat[max_order::] = y[max_order::] - np.dot(X, beta_hat)
            # adding non-identified outputs
            Yid = np.hstack((y[:max_order], np.dot(X, beta_hat)))

        if iterations >= max_iter:
            warn("[ARMAX_id] Reached maximum iterations.")
            # max_reached = True

        numerator = np.zeros(max_order)
        numerator[theta : nb + theta] = beta_hat[na : na + nb]

        denominator = np.zeros(max_order + 1)
        denominator[0] = 1.0
        denominator[1 : na + 1] = beta_hat[0:na]

        numerator_H = np.zeros(max_order + 1)
        numerator_H[0] = 1.0
        numerator_H[1 : nc + 1] = beta_hat[na + nb : :]

        denominator_H = np.zeros(max_order + 1)
        denominator_H[0] = 1.0
        denominator_H[1 : na + 1] = beta_hat[0:na]

        return numerator, denominator, numerator_H, denominator_H, Vn, Yid
